import os
import re
import json
import arcgis
import logging
import warnings
import traceback
from datetime import datetime

HAS_NUMPY = True
HAS_FASTAI = True
HAS_TRANSFORMER = True
HAS_BEAUTIFULSOUP = True


try:
    import torch
    import pandas as pd
    from PIL import Image as PIL_Image
    from pathlib import Path
    from arcgis.auth.tools import LazyLoader

    cv2 = LazyLoader("cv2")
    h5py = LazyLoader("h5py")
    px = LazyLoader("plotly.express")
    go = LazyLoader("plotly.graph_objs")
    from torchvision import models
    from sklearn.cluster import DBSCAN, KMeans
    from sklearn.decomposition import PCA
    from fastai.vision.data import imagenet_stats
    from fastprogress.fastprogress import progress_bar
    from .._data import _make_folder
    from ..models._arcgis_model import (
        _device_check,
        _resnet_family,
        _vgg_family,
        _densenet_family,
    )
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False
else:
    warnings.filterwarnings("ignore", category=UserWarning, module="fastai")

try:
    import numpy as np
except Exception as e:
    HAS_NUMPY = False
else:
    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)

try:
    from bs4 import BeautifulSoup
except Exception as e:
    HAS_BEAUTIFULSOUP = False
else:
    warnings.filterwarnings("ignore", category=UserWarning, module="bs4")

max_token_length = 512
allowed_text_extensions = ["csv", "txt", "json"]
allowed_image_extensions = ["png", "jpg", "jpeg", "tiff", "tif", "bmp"]


class TextModule:
    @staticmethod
    def preprocess_text(text, remove_urls=False, remove_html_tags=False):
        """
        Perform some basic cleanup like removing HTML tags, removing urls,
        and converting multiple white spaces to single white space
        """
        if remove_urls:
            text = re.sub(
                r"\b(?:(?:https?|ftp)://)?\w[\w-]*(?:\.[\w-]+)+\S*", " ", text
            )

        # text = re.sub(r'<.*?>', '', text)
        if remove_html_tags:
            if not HAS_BEAUTIFULSOUP:
                raise Exception("This module requires BeautifulSoup.")
            text = BeautifulSoup(text, "html.parser").get_text(
                separator=" ", strip=True
            )

        if any([remove_html_tags, remove_urls]):
            text = re.sub(" +", " ", text)
        return text.strip()

    @classmethod
    def preprocess_text_list(cls, text_list, remove_urls=False, remove_html_tags=False):
        text_list = [
            cls.preprocess_text(x, remove_urls, remove_html_tags) for x in text_list
        ]
        return text_list

    # Mean Pooling - Take attention mask into account for correct averaging
    @staticmethod
    def mean_pooling(model_output, attention_mask):
        token_embeddings = model_output
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)
        sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
        return sum_embeddings / sum_mask

    # Max Pooling - Take the max value over time for every dimension
    @staticmethod
    def max_pooling(model_output, attention_mask):
        token_embeddings = model_output
        input_mask_expanded = (
            attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        )
        token_embeddings[
            input_mask_expanded == 0
        ] = -1e9  # Set padding tokens to large negative value
        max_over_time = torch.max(token_embeddings, 1)[0]
        return max_over_time

    # Take the first token ([CLS]) from each sentence
    @staticmethod
    def cls_token(model_output, mask=None):
        return model_output[:, 0]


class Embeddings:
    """
    Creates an :class:`~arcgis.learn.Embeddings` Object. This object is capable of giving
    embeddings for text as well as images. The image embeddings are
    currently supported for RGB images only

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    dataset_type            Required string. The type of data for which
                            we would like to get the embedding vectors.
                            Valid values are `text` & `image`. Default
                            is set to `image`.

                            .. note::
                                The image embeddings are currently supported for `RGB` images only.
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the backbone/model-name
                            to be used to get the embedding vectors.
                            Default backbone for `image` dataset-type is
                            `resnet34` and for `text` dataset-type is
                            `sentence-transformers/distilbert-base-nli-stsb-mean-tokens`

                            To learn more about the available models for
                            for getting `text` embeddings, kindly visit:-
                            https://huggingface.co/sentence-transformers
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.Embeddings` Object
    """

    def __init__(self, dataset_type="image", backbone=None, **kwargs):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)
        self._dataset_type = dataset_type

        if self._dataset_type == "image":
            self._allowed_extensions = allowed_image_extensions
        elif self._dataset_type == "text":
            self._allowed_extensions = allowed_text_extensions

        if "working_dir" in kwargs:
            self.working_dir = kwargs.get("working_dir")
        else:
            self.working_dir = Path.cwd()

        # _make_folder(os.path.join(os.path.abspath(self.working_dir), "embeddings"))

        self._file_path = None
        self.backbone = None
        self._tokenizer = None
        self._device = self._get_device()
        self._error_message = (
            f"Wrong backbone - {backbone} choosen for dataset-type - {self._dataset_type}. Kindly "
            f"call the `Embeddings.supported_backbones('{self._dataset_type}')` to see the "
            f"supported backbones for - {self._dataset_type} dataset-type."
        )

        from arcgis._impl.common._utils import _DisableLogger

        with _DisableLogger():
            self.model = self._load_model(dataset_type, backbone)
        self.model.to(self._device)

    @classmethod
    def supported_backbones(cls, dataset_type="image"):
        """
        Get available backbones/model-name for the given `dataset-type`

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        dataset_type            Required string. The type of data for which
                                we would like to get the embedding vectors.
                                Valid values are `text` & `image`. Default
                                is set to `image`
        =====================   ===========================================

        :return: a list containing the available models for the given `dataset-type`
        """
        if dataset_type == "image":
            return cls._get_image_compatible_backbones()
        elif dataset_type == "text":
            return cls._get_text_compatible_backbones()
        else:
            error_message = f"Wrong dataset-type - {dataset_type} provided. Valid values are 'image' or 'text'."
            raise Exception(error_message)

    @staticmethod
    def _get_image_compatible_backbones():
        return [
            *_resnet_family,
            *_densenet_family,
            *_vgg_family,
            models.mobilenet_v2.__name__,
        ]

    @staticmethod
    def _get_text_compatible_backbones():
        return [
            "sentence-transformers/distilbert-base-nli-stsb-mean-tokens",
            "sentence-transformers/bert-base-nli-max-tokens",
            "sentence-transformers/bert-base-nli-cls-token",
        ] + [
            "See all `TextEmbedding` models at https://huggingface.co/sentence-transformers"
        ]

    def _get_device(self):
        move_to_cpu = _device_check()
        if move_to_cpu:
            arcgis.env._processorType = "CPU"

        if (
            getattr(arcgis.env, "_processorType", "") == "GPU"
            and torch.cuda.is_available()
        ):
            device = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            device = torch.device("cpu")
        else:
            device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )
        return device

    def _load_model(self, dataset_type="image", backbone=None):
        if dataset_type == "image":
            model = self._load_image_model(backbone)
        elif dataset_type == "text":
            model = self._load_text_model(backbone)
        else:
            error_message = f"Wrong dataset-type - {dataset_type} provided. Valid values are 'image' or 'text'."
            raise Exception(error_message)
        return model

    def _load_image_model(self, backbone=None):
        if backbone is None:
            backbone = "resnet34"

        if hasattr(models, backbone):
            model_type = getattr(models, backbone)
        else:
            raise Exception(self._error_message)

        self.backbone = backbone
        model = model_type(pretrained=True, progress=True)
        if "vgg" in self.backbone:
            model.classifier = model.classifier[:-1]
        else:
            layers = list(model.children())[:-1]
            model = torch.nn.Sequential(*layers)

        return model.eval()

    def _load_text_model(self, backbone=None):
        HAS_TRANSFORMER = True
        try:
            from transformers import AutoTokenizer, AutoModel
        except Exception as e:
            HAS_TRANSFORMER = False
        if not HAS_TRANSFORMER:
            raise Exception("This module requires transformers library.")
        if backbone is None:
            backbone = "sentence-transformers/distilbert-base-nli-stsb-mean-tokens"
        self.backbone = backbone
        try:
            model = AutoModel.from_pretrained(backbone)
            self._tokenizer = AutoTokenizer.from_pretrained(
                backbone, config=model.config
            )
        except Exception as e:
            raise Exception(self._error_message)

        return model.eval()

    def get(
        self,
        text_or_list,
        batch_size=32,
        show_progress=True,
        return_embeddings=False,
        **kwargs,
    ):
        """
        Method to get the embedding vectors for the image/text items.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or List. String containing
                                directory path or list of directory paths where
                                image/text files are present for which the user wants
                                to get the embedding vectors.
        ---------------------   -------------------------------------------
        batch_size              Optional integer. The number of items to process
                                in one batch. Default is set to 32.
        ---------------------   -------------------------------------------
        show_progress           Optional boolean. If set to True, will display a
                                progress bar depicting the items processed so far.
                                Default is set to `True`.
        ---------------------   -------------------------------------------
        return_embeddings       Optional boolean. If set to True, a dataframe
                                containing the embeddings will be returned. If set
                                to False, they will be saved in a h5 file.
                                Default is set to `False`.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        normalize               Optional boolean. If set to `true`, will normalize
                                the image with `imagenet-stats` (mean and
                                std-deviation for each color channel in RGB image).
                                This argument is valid only for `dataset-type` image.
                                Default is set to True.
        ---------------------   -------------------------------------------
        file_extensions         Optional String or List. The file extension(s) for
                                which the user wish to get embedding vectors for.
                                Allowed values for `dataset-type` image are -
                                ['png', 'jpg', 'jpeg', 'tiff', 'tif', 'bmp']
                                Allowed values for `dataset-type` text are -
                                ['csv', 'txt', 'json']

                                .. note::
                                        For json files, if we have nested json structures, then text will be extracted only from the 1st level.
        ---------------------   -------------------------------------------
        chip_size               Optional integer. Resize the image to
                                `chip_size X chip_size` pixels.
                                This argument is valid only for `dataset-type` image.
                                Default is set to 224
        ---------------------   -------------------------------------------
        encoding                Optional string. The encoding to read the text/csv/
                                json file. Applicable only for `dataset-type` text.
                                Default is `UTF-8`
        ---------------------   -------------------------------------------
        text_column             Optional string. The column that will be used to get
                                the text content from `csv` or `json` file types.
                                This argument is valid only for `dataset-type` text.
                                Default is set to `text`
        ---------------------   -------------------------------------------
        remove_urls             Optional boolean. If true, remove urls from text.
                                This argument is valid only for `dataset-type` text.
                                Default value is False.
        ---------------------   -------------------------------------------
        remove_html_tags        Optional boolean. If true, remove html tags from text.
                                This argument is valid only for `dataset-type` text.
                                Default value is False.
        ---------------------   -------------------------------------------
        pooling_strategy        Optional string. The transformer model gives embeddings
                                for each word/token present in the text. The type of
                                pooling to be done on those word/token vectors in order
                                to form the text embeddings.
                                Allowed values are - ['mean', 'max', 'first']
                                This argument is valid only for `dataset-type` text.
                                Default value is `mean`.
        =====================   ===========================================

        :return: The path of the H5 file where items & corresponding embeddings are saved.
        """
        if not HAS_NUMPY:
            raise Exception("This module requires numpy.")

        file_name = f"embeddings_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.h5"
        self._file_path = os.path.join(self.working_dir, "embeddings", file_name)
        if os.path.exists(self._file_path):
            raise Exception(
                f"File to save the embeddings already present at - {self._file_path}. Kindly rename the file "
                f"or move the file to another location to proceed."
            )
        text_img_df = kwargs.get("dataframe", False)
        if isinstance(text_img_df, pd.DataFrame):
            col = kwargs.get("text_column", "text")
            item_list = text_img_df[col].values.tolist()
        else:
            item_list = self._get_items(text_or_list, **kwargs)
        if self._dataset_type == "image":
            ret = self._get_image(
                item_list, batch_size, show_progress, return_embeddings, **kwargs
            )
        else:
            ret = self._get_text(
                item_list, batch_size, show_progress, return_embeddings, **kwargs
            )

        return ret

    @staticmethod
    def _do_h5file_sanity(file_path):
        hf = h5py.File(file_path, "r")
        dataset_keys = list(hf.keys())
        if (
            len(dataset_keys) != 2
            or "embeddings" not in dataset_keys
            or "items" not in dataset_keys
        ):
            error_message = (
                f"Wrong H5 file passed. The H5 files consists of - {dataset_keys} dataset keys. "
                f"We expected the following dataset keys - ['items', 'embeddings']"
            )
            hf.close()
            raise Exception(error_message)

        embeddings_dataset, items_dataset = hf.get("embeddings"), hf.get("items")
        if embeddings_dataset.shape[0] != items_dataset.shape[0]:
            error_message = (
                f"Length of `embeddings`({embeddings_dataset.shape[0]}) and `items`"
                f"({items_dataset.shape[0]}) do not match. There's some problem with the "
                f"file present at location - {file_path}"
            )
            raise Exception(error_message)

        hf.close()

    def load(self, file_path, load_to_memory=True):
        """
        Load the extracted embeddings from the H5 file

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        file_path               Required string. The path to the H5 file which
                                gets auto generated after the call to the `get`
                                method of the :class:`~arcgis.learn.Embeddings` class
        ---------------------   -------------------------------------------
        load_to_memory          Optional Bool. whether or not to load the entire
                                content of the H5 file to memory. Loading very large
                                H5 files into the memory takes up lot of RAM space.
                                Use this parameter with caution for large H5 files.
                                Default is set to True.
        =====================   ===========================================

        :return: When `load_to_memory` param is `True` - A 2 item tuple containing
                  the numpy arrays of extracted embeddings and items
                  When `load_to_memory` param is `False` - A 3 item tuple containing
                  the H5 file handler & 2 H5 dataset object of extracted embeddings
                  and items
        """

        if os.path.exists(file_path) is False or os.path.isfile(file_path) is False:
            raise Exception(f"The path `{file_path}` does not exists, or not a file")

        self._do_h5file_sanity(file_path)
        hf = h5py.File(file_path, "r")
        embeddings_dataset, items_dataset = hf.get("embeddings"), hf.get("items")
        if load_to_memory:
            embeddings, items = np.array(embeddings_dataset), np.array(items_dataset)
            hf.close()
            return embeddings, items
        else:
            return hf, embeddings_dataset, items_dataset

    @staticmethod
    def _check_directory_validity(dir_paths):
        for directory in dir_paths:
            if os.path.exists(directory) is False:
                raise Exception(f"The path `{directory}` does not exists.")
            if os.path.isfile(directory):
                raise Exception(
                    f"The path `{directory}` seems like a file path. Please provide the directory path."
                )
            if os.path.isdir(directory) is False:
                raise Exception(
                    f"The path `{directory}` is not a valid directory path."
                )

    def _check_file_extension_validity(self, file_extensions):
        for extension in file_extensions:
            if extension not in self._allowed_extensions:
                raise Exception(
                    f"Extension -`{extension}` is not a valid extension for dataset-type - "
                    f"`{self._dataset_type}`. Allowed extension values are - {self._allowed_extensions}"
                )

    def _get_text_items(self, file_paths, text_column, encoding):
        text_list = []
        for file_path, ext in file_paths:
            if ext == "txt":
                with open(file_path, "r", encoding=encoding, errors="ignore") as f:
                    text_list.append(f.read())
            elif ext == "csv":
                df = pd.read_csv(file_path, dtype="str")
                if text_column not in df.columns:
                    raise Exception(
                        f"CSV file - {file_path} doesn't contain the column - `{text_column}`"
                    )
                df.dropna(axis=0, subset=[text_column], inplace=True)
                text_list.extend(df[text_column].values.tolist())
            elif ext == "json":
                with open(file_path, "r", encoding=encoding) as f:
                    data_list = f.readlines()

                txt_list = []
                for item in data_list:
                    data_dict = json.loads(item)
                    if text_column not in data_dict:
                        raise Exception(
                            f"Row - \n{data_dict}\n\n in JSON file - {file_path} doesn't contain the key - "
                            f"`{text_column}`. Kindly fix the JSON file or pass the correct value of the "
                            f"`text_column` parameter in the `get` method call."
                        )
                    text = data_dict[text_column]
                    if text:
                        txt_list.append(text)

                text_list.extend(txt_list)
            else:
                raise Exception(
                    f"Extension -`{ext}` is not a valid extension for dataset-type - `{self._dataset_type}`"
                    f". Allowed extension values are - {self._allowed_extensions}"
                )

        return text_list

    def _get_items(self, dir_path, **kwargs):
        item_list = []
        file_extensions = kwargs.get("file_extensions", self._allowed_extensions)
        if isinstance(file_extensions, (str, bytes)):
            file_extensions = [file_extensions]
        file_extensions = [x.replace(".", "") for x in file_extensions]

        self._check_file_extension_validity(file_extensions)
        text_column = kwargs.get("text_column", "text")
        encoding = kwargs.get("encoding", "utf-8")
        if isinstance(dir_path, (str, bytes)):
            dir_path = [dir_path]
        self._check_directory_validity(dir_path)

        for rootdir in dir_path:
            for dirpath, dirs, files in os.walk(rootdir):
                for filename in files:
                    fname = os.path.join(dirpath, filename)
                    ext = os.path.splitext(filename)[-1].lower().replace(".", "")
                    if ext in file_extensions:
                        logging.info(
                            f"File name - {fname} will be considered for getting the embeddings"
                        )
                        item_list.append((fname, ext))

        if len(item_list) == 0:
            raise Exception(
                f"Not a single item found to extract the embeddings in folder(s) - {dir_path}. Kindly check"
                f" if the directory contains image/text files or check if you have passed the right set of"
                f" `file_extensions` to the method."
            )

        if self._dataset_type == "text":
            item_list = self._get_text_items(item_list, text_column, encoding)
        else:
            item_list = [x[0] for x in item_list]

        return item_list

    @staticmethod
    def _normalize(x, mean, std):
        z = (x - mean[..., None, None]) / std[..., None, None]
        return z

    @staticmethod
    def _insert_to_h5_file(file_handler, items, embeddings):
        dt = h5py.special_dtype(vlen=bytes)
        if len(file_handler.keys()) == 0:
            file_handler.create_dataset(
                "items", data=items, maxshape=(None,), dtype=dt, chunks=True
            )
            file_handler.create_dataset(
                "embeddings",
                data=embeddings,
                maxshape=(None, embeddings.shape[1]),
                chunks=True,
            )
        else:
            file_handler["items"].resize(
                (file_handler["items"].shape[0] + items.shape[0]), axis=0
            )
            file_handler["items"][-items.shape[0] :] = items
            file_handler["embeddings"].resize(
                (file_handler["embeddings"].shape[0] + embeddings.shape[0]), axis=0
            )
            file_handler["embeddings"][-embeddings.shape[0] :] = embeddings

    def _get_image(
        self,
        item_list,
        batch_size=32,
        show_progress=True,
        return_embeddings=False,
        **kwargs,
    ):
        if not return_embeddings:
            with h5py.File(self._file_path, "a") as hf:
                batch_embeddings = self._extract_img_embeddings(
                    item_list, batch_size=32, show_progress=True, **kwargs
                )
                self._insert_to_h5_file(hf, item_list, batch_embeddings)
            return self._file_path
        else:
            batch_embeddings = self._extract_img_embeddings(
                item_list, batch_size=32, show_progress=True, **kwargs
            )
            return batch_embeddings

    def _extract_img_embeddings(
        self, item_list, batch_size=32, show_progress=True, **kwargs
    ):
        normalize = kwargs.get("normalize", True)
        resize_to = kwargs.get("chip_size", 224)
        all_batch_embedding = np.empty([0, 512])
        mean, std = None, None
        if normalize:
            mean, std = imagenet_stats
            mean = torch.tensor(mean).float().to(self._device)
            std = torch.tensor(std).float().to(self._device)

        for i in progress_bar(
            range(0, len(item_list), batch_size), display=show_progress
        ):
            try:
                img_list = item_list[i : i + batch_size]
                img_batch = np.array(
                    [
                        np.array(
                            PIL_Image.open(img_path)
                            .convert("RGB")
                            .resize((resize_to, resize_to))
                        ).astype(float)
                        / 255.0
                        for img_path in img_list
                    ]
                )

                img_batch = (
                    torch.tensor(img_batch.transpose(0, 3, 1, 2))
                    .float()
                    .to(self._device)
                )
                if normalize:
                    img_batch = self._normalize(img_batch, mean, std)

                with torch.no_grad():
                    out = self.model(img_batch)
                    if "densenet" in self.backbone or "mobilenet" in self.backbone:
                        out = torch.nn.AdaptiveAvgPool2d(output_size=(1, 1))(out)
                    if (
                        "densenet" in self.backbone
                        or "mobilenet" in self.backbone
                        or "resnet" in self.backbone
                    ):
                        out = out.view(out.size(0), -1)

                img_list = np.array([x.encode() for x in img_list])
                batch_embeddings = (
                    torch.nn.functional.normalize(out.data).cpu().detach().numpy()
                )
                all_batch_embedding = np.append(
                    all_batch_embedding, batch_embeddings, axis=0
                )

                # self._insert_to_h5_file(hf, img_list, batch_embeddings)
            except Exception as e:
                raise Exception(e)
        return all_batch_embedding

    def _get_text(
        self,
        item_list,
        batch_size=32,
        show_progress=True,
        return_embeddings=False,
        **kwargs,
    ):
        if not return_embeddings:
            with h5py.File(self._file_path, "a") as hf:
                batch_embeddings = self._extract_text_embeddings(
                    item_list, batch_size=32, show_progress=True, **kwargs
                )
                self._insert_to_h5_file(hf, img_list, batch_embeddings)
            return self._file_path
        else:
            batch_embeddings = self._extract_text_embeddings(
                item_list, batch_size=32, show_progress=True, **kwargs
            )
            return batch_embeddings

    def _extract_text_embeddings(
        self, item_list, batch_size=32, show_progress=True, **kwargs
    ):
        remove_urls = kwargs.get("remove_urls", False)
        remove_html_tags = kwargs.get("remove_html_tags", False)
        pooling_strategy = kwargs.get("pooling_strategy", "mean")

        all_batch_embedding = np.empty([0, 768])
        if any([remove_urls, remove_html_tags]):
            item_list = TextModule.preprocess_text_list(
                item_list, remove_urls, remove_html_tags
            )
        # batch_embedding_list = []
        try:
            for i in progress_bar(
                range(0, len(item_list), batch_size), display=show_progress
            ):
                text_batch = item_list[i : i + batch_size]
                encoded_input = self._tokenizer(
                    text_batch,
                    padding=True,
                    truncation=True,
                    max_length=max_token_length,
                    return_tensors="pt",
                ).to(self._device)
                with torch.no_grad():
                    model_output = self.model(**encoded_input)

                # First element of model_output contains all token embeddings
                # token_embeddings = torch.nn.functional.normalize(model_output[0])
                token_embeddings, attention_mask = (
                    model_output[0],
                    encoded_input["attention_mask"],
                )
                if pooling_strategy == "mean":
                    batch_embedding = TextModule.mean_pooling(
                        token_embeddings, attention_mask
                    )
                elif pooling_strategy == "max":
                    batch_embedding = TextModule.max_pooling(
                        token_embeddings, attention_mask
                    )
                elif pooling_strategy == "first":
                    batch_embedding = TextModule.cls_token(token_embeddings)
                else:
                    error_message = (
                        f"Wrong pooling-strategy - {pooling_strategy} choosen. Allowed values are - "
                        f"'mean', 'max' and 'first'. kindly choose from these options."
                    )
                    raise Exception(error_message)

                text_batch = np.array([x.encode() for x in text_batch])
                batch_embeddings = (
                    torch.nn.functional.normalize(batch_embedding)
                    .cpu()
                    .detach()
                    .numpy()
                )
                all_batch_embedding = np.append(
                    all_batch_embedding, batch_embeddings, axis=0
                )

        except Exception as e:
            raise Exception(e)
        return all_batch_embedding

    @staticmethod
    def _do_clustering(embeddings, item_list=None, n_clusters=5, dimensions=3):
        n_components = min(embeddings.shape[0], 64)
        transformed_embeddings = PCA(n_components=n_components).fit_transform(
            embeddings
        )
        embeddings_for_visualization = PCA(n_components=dimensions).fit_transform(
            embeddings
        )

        columns = ["x", "y", "z"] if dimensions == 3 else ["x", "y"]
        result = pd.DataFrame(embeddings_for_visualization, columns=columns)
        if item_list:
            result["item"] = item_list

        # DBSCAN clustering
        # eps = kwargs.pop("eps", 0.5)
        # min_samples, metric = kwargs.pop("min_samples", 10), kwargs.pop("metric", "cosine")
        # clustering = DBSCAN(eps=eps, min_samples=min_samples, metric=metric, **kwargs).fit(transformed_embeddings)
        # result['labels'] = clustering.labels_
        # outliers = result.loc[result.labels == -1, :]
        # clustered = result.loc[result.labels != -1, :]

        # KMeans clustering
        clustering = KMeans(n_clusters=n_clusters)
        labels = clustering.fit_predict(transformed_embeddings)
        result["labels"] = labels
        random_sample = result.reset_index(drop=True)

        if len(random_sample) > 500:
            num_items_per_group = int(500 / n_clusters)
            # random_sample = random_sample.sample(n=500).reset_index(drop=True)
            random_sample = (
                random_sample.groupby("labels")
                .apply(lambda x: x.sample(n=num_items_per_group, replace=True))
                .reset_index(drop=True)
            )

        # print(len(random_sample))
        return random_sample

    def _visualize_with_items(self, cluster_dataframe, dimensions=3):
        from ipywidgets import Image, Layout, HBox, Textarea

        image_data = {}
        widget_dict = dict(
            x=cluster_dataframe["x"],
            y=cluster_dataframe["y"],
            mode="markers",
            marker=dict(color=cluster_dataframe["labels"]),
        )
        if dimensions == 3:
            widget_dict.update({"z": cluster_dataframe["z"], "type": "scatter3d"})
        else:
            widget_dict.update({"type": "scatter"})

        fig = go.FigureWidget(data=[widget_dict])

        scatter = fig.data[0]

        if self._dataset_type == "image":
            for img_path in cluster_dataframe["item"]:
                img = np.array(PIL_Image.open(img_path))
                inmem_jpg = cv2.imencode(".png", cv2.cvtColor(img, cv2.COLOR_BGR2RGB))[
                    1
                ].tobytes()
                image_data[img_path] = inmem_jpg

            widget = Image(
                value=image_data[cluster_dataframe.item.iloc[0]],
                layout=Layout(height="200px", width="200px"),
            )
        else:
            widget = Textarea(
                value=cluster_dataframe.item.iloc[0],
                disabled=True,
                layout=Layout(height="350px", width="200px"),
            )

        def hover_fn(trace, points, state):
            ind = points.point_inds[0]
            item = cluster_dataframe["item"][ind]
            widget.value = image_data[item] if self._dataset_type == "image" else item

        scatter.on_hover(hover_fn)
        return HBox([fig, widget])

    def visualize(
        self, file_path, visualize_with_items=True, n_clusters=5, dimensions=3
    ):
        """
        Method to visualize the embedding vectors for the image/text items.
        This method uses the K-Means clustering algorithm to partition the
        embeddings vectors into n-clusters. This requires the loading the
        entire content of the H5 file to RAM. Loading very large H5 files
        into the memory takes up lot of RAM space. Use this method with
        caution for large H5 files.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        file_path               Required string. The path to the H5 file which
                                gets auto generated after the call to the `get`
                                method of the :class:`~arcgis.learn.Embeddings` class.
        ---------------------   -------------------------------------------
        visualize_with_items    Optional Bool. Whether or not to visualize the
                                embeddings with items. Default is set to True.
        ---------------------   -------------------------------------------
        n_clusters              Optional integer. The number of clusters to create
                                for the embedding vectors. This value will be passed
                                to the `KMeans` algorithm to generate the clusters.
                                Default is set to 5.
        ---------------------   -------------------------------------------
        dimensions              Optional integer. The number of dimensions to project
                                the embedding vectors for visualization purpose.
                                Allowed values are `2` & `3`
                                Default is set to 3.
        =====================   ===========================================
        """

        self._do_h5file_sanity(file_path)
        hf = h5py.File(file_path, "r")
        embeddings_dataset, items_dataset = hf.get("embeddings"), hf.get("items")
        embeddings, item_list = (
            np.array(embeddings_dataset),
            np.array(items_dataset).tolist(),
        )
        hf.close()

        # For using DBSCAN clustering take `eps`, `metric` and `min_samples` in **kwargs and pass to below method
        cluster_df = self._do_clustering(
            embeddings, item_list, n_clusters=n_clusters, dimensions=dimensions
        )

        if visualize_with_items is False:
            if dimensions == 3:
                fig = px.scatter_3d(
                    cluster_df,
                    x=cluster_df.x,
                    y=cluster_df.y,
                    z=cluster_df.z,
                    color=cluster_df.labels,
                    hover_data=["labels"],
                )
            else:
                fig = px.scatter(
                    cluster_df,
                    x=cluster_df.x,
                    y=cluster_df.y,
                    color="labels",
                    hover_data=["labels"],
                )
            return fig
        else:
            return self._visualize_with_items(cluster_df, dimensions)



from __future__ import annotations
import json
from collections import OrderedDict
from urllib.request import HTTPError
from arcgis.gis._impl._con import Connection
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
from typing import Dict, Any, Optional, List


###########################################################################
class KubeEnterpriseGroups:
    """
    This resource is an umbrella for operations that inventory your
    organization's groups, such as retrieving a list of users within a
    specific group or listing which groups a specific user is assigned to.
    The groups resource returns the total number of enterprise groups in
    the system.
    """

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS"):
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        returns the properties for the Organization

        :return: dict
        """
        if self._properties is None:
            self._properties = self._con.get(self._url, {"f": "json"})
        return self._properties

    # ----------------------------------------------------------------------
    def search(self, query: str = "", max_count: int = 1000) -> dict:
        """
        Searches users in the configured enterprise user store.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        query                           Optional String. Text to narrow down the user search.
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional Integer.  The maximum number of recrods that the client will accept.
        ===========================     ====================================================================

        """
        url = f"{self._url}/searchEnterpriseGroups"
        params = {"f": "json", "filter": query, "maxCount": max_count}
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def find_within_groups(self, name: str, query: str = None, max_count: int = 1000):
        """
        This operation returns a list of users that are currently assigned to the enterprise group within the enterprise user and group stores.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        name                            Required String. The name of the group.
        ---------------------------     --------------------------------------------------------------------
        query                           Optional String. Text to narrow down the user search.
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional Integer.  The maximum number of recrods that the client will accept.
        ===========================     ====================================================================
        """
        if query is None:
            query = ""
        url = f"{self._url}/getUsersWithinEnterpriseGroup"
        params = {
            "f": "json",
            "groupName": name,
            "filter": query,
            "maxCount": max_count,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def get_user_groups(
        self, username: str, query: str = None, max_count: int = 1000
    ) -> dict:
        """
        This operation searches groups in the configured role store. You can narrow down the search using the `query` parameter.


        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required String. The username to examine.
        ---------------------------     --------------------------------------------------------------------
        query                           Optional String. Text to narrow down the user search.
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional Integer.  The maximum number of recrods that the client will accept.
        ===========================     ====================================================================

        :returns: dict

        """
        if query is None:
            query = ""

        url = f"{self._url}/getEnterpriseGroupsForUser"
        params = {
            "f": "json",
            "username": username,
            "filter": query,
            "maxCount": max_count,
        }
        return self._con.post(url, params)

    # ----------------------------------------------------------------------
    def refresh_membership(self, groups: List[str]) -> bool:
        """
        This operation iterates over every enterprise account configured in
        your organization and determines whether the user account is part
        of the input enterprise group. If there are any changes in
        membership, the database and indexes are updated for each group.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        groups                          Required List[str]. The name of the groups to refresh.
        ===========================     ====================================================================

        :returns: bool
        """
        assert isinstance(groups, (list, tuple))
        groups = ",".join(groups)
        url = f"{self._url}/refreshMembership"
        params = {"groups": groups, "f": "json"}
        return self._con.post(url, params).get("status", False)


###########################################################################
class KubeOrgSecurity(object):
    """
    Allows the for the management of the security of the settings.
    """

    _con = None
    _gis = None
    _url = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS") -> "KubeOrgSecurity":
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        returns the properties for the Organization

        :return: dict
        """
        if self._properties is None:
            self._properties = self._con.get(self._url, {"f": "json"})
        return self._properties

    @property
    def enterprise_user(self):
        """Allows users to manage and work with enterprise users"""
        url = f"{self._url}/users"
        return KubeEnterpriseUser(url, gis=self._gis)

    @property
    def enterprise_groups(self):
        """Allows users to manage and work with enterprise groups"""
        url = f"{self._url}/groups"
        return KubeEnterpriseGroups(url, gis=self._gis)


class KubeEnterpriseUser:
    """
    The `KubeEnterpriseUser` resource houses operations used to manage
    members in your organization.
    """

    _url = None
    _gis = None

    def __init__(self, url, gis):
        self._url = url
        self._gis = gis

    def search(self, query: str = "", max_count: int = 1000) -> dict:
        """
        Searches users in the configured enterprise user store.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        query                           Optional String. Text to narrow down the user search.
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional Integer.  The maximum number of recrods that the client will accept.
        ===========================     ====================================================================

        """
        url = f"{self._url}/searchEnterpriseUsers"
        params = {"f": "json", "filter": query, "maxCount": max_count}
        return self._gis._portal.con.post(url, params)

    def create_user(
        self,
        username: str,
        password: str,
        first_name: str,
        last_name: str,
        email: str,
        user_license: str,
        role: str = "org_user",
        provider: str = "arcgis",
        idp_username: str | None = None,
        description: str | None = None,
    ) -> bool:
        """
        This operation is used to pre-create built-in or enterprise
        accounts within the portal. The provider parameter is used to
        indicate the type of user account.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The name of the user account
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password of the user account
        ---------------------------     --------------------------------------------------------------------
        first_name                      Required string. The first name for the account
        ---------------------------     --------------------------------------------------------------------
        last_name                       Required string. The last name for the account
        ---------------------------     --------------------------------------------------------------------
        email                           Required string. The email for the account
        ---------------------------     --------------------------------------------------------------------
        user_license	                Optional string. The user type for the account.

                                        Values: creator, editor, advanced (GIS Advanced),
                                                basic (GIS Basic), standard (GIS Standard), viewer,
                                                fieldworker
        ---------------------------     --------------------------------------------------------------------
        role                            Optional string. The role for the user account. The default value is
                                        org_user.
                                        Values org_admin | org_publisher | org_user | org_editor (Data Editor) | viewer
        ---------------------------     --------------------------------------------------------------------
        provider                        Optional string. The provider for the account. The default value is
                                        arcgis. Values arcgis | enterprise
        ---------------------------     --------------------------------------------------------------------
        idp_username                    Optional string. The name of the user as stored by the enterprise
                                        user store. This parameter is only required if the provider
                                        parameter is enterprise.
        ---------------------------     --------------------------------------------------------------------
        description                     Optional string. A user description
        ===========================     ====================================================================

        :return: boolean

        """
        role_lu = {
            "editor": "iBBBBBBBBBBBBBBB",
            "viewer": "iAAAAAAAAAAAAAAA",
            "org_editor": "iBBBBBBBBBBBBBBB",
            "org_viewer": "iAAAAAAAAAAAAAAA",
        }
        user_license_lu = {
            "creator": "creatorUT",
            "editor": "editorUT",
            "advanced": "GISProfessionalAdvUT",
            "basic": "GISProfessionalBasicUT",
            "standard": "GISProfessionalStdUT",
            "viewer": "viewerUT",
            "fieldworker": "fieldWorkerUT",
        }
        if user_license.lower() in user_license_lu:
            user_license = user_license_lu[user_license.lower()]
        if role.lower() in role_lu:
            role = role_lu[role.lower()]

        url = "%s/createUser" % self._url
        params = {
            "f": "json",
            "username": username,
            "password": password,
            "firstname": first_name,
            "lastname": last_name,
            "email": email,
            "role": role,
            "provider": provider,
        }
        if idp_username:
            params["idpUsername"] = idp_username
        if description:
            params["description"] = description
        if user_license:
            params["userLicenseTypeId"] = user_license
        res = self._gis._portal.con.post(path=url, postdata=params)
        return res["status"] == "success"

    def get_enterprise_user(self, username: str) -> dict:
        """gets the enterprise user"""
        url = f"{self._url}/getEnterpriseUser"
        params = {"f": "json", "username": username}
        return self._gis._portal.con.post(url, params)

    def refresh_membership(self, users: list[str]) -> dict:
        """
        This operation iterates over every enterprise group configured in
        your organization and determines whether the input user accounts
        belong to any of the configured enterprise groups. If there is any
        change in membership, the database and the indexes are updated for
        each user account. While the portal automatically refreshes the
        memberships during a user login and during a periodic refresh
        (configured through the Update Identity Store operation), this
        operation allows an administrator to force a refresh.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        users                           Optional list[str]. The comma-separated list of usernames for
                                        whom the memberships need to be refreshed.
        ===========================     ====================================================================

        :returns: dict
        """
        if isinstance(users, (list, tuple)):
            users = ",".join([str(user) for user in users])
        url = f"{self._url}/refreshMembership"
        params = {"f": "json", "users": users}
        return self._gis._portal.con.post(url, params)


###########################################################################
class KubeOrganization:
    """
    A single organization within your deployment, allowing you to manage
    and update it's licensing and security information, as well as manage
    it's federated servers.
    """

    _con = None
    _gis = None
    _url = None
    _properties = None
    _security = None
    _federation = None
    _license = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis: "GIS", **kwargs):
        """class initializer"""
        self._gis = gis
        self._url = url
        self._con = gis._con
        self._properties = None
        self._json_dict = None

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties into the class"""
        params = {"f": "json"}
        try:
            result = self._con.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def org_property(self) -> dict:
        """
        This operation lists and sets properties specific to an organization that
        can be modified to control your deployment.

        :Returns: dict

        """
        url = f"{self._url}/properties"
        params = {"f": "json"}
        res = self._con.get(url, params)
        return res

    # ----------------------------------------------------------------------
    @org_property.setter
    def org_property(self, value: dict):
        """
        This operation lists and sets properties specific to an organization that
        can be modified to control your deployment.

        :Returns: dict
        """
        url = f"{self._url}/properties/update"
        params = {"f": "json"}
        res = self._con.get(url, params)
        if res.get("status") == False:
            raise Exception(res)

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def security(self):
        if self._security is None:
            self._security = KubeOrgSecurity(url=f"{self._url}/security", gis=self._gis)
        return self._security

    # ----------------------------------------------------------------------
    @property
    def license(self) -> "KubeOrgLicense":
        """
        The Licenses resource returns high-level licensing details.

        :return: KubeOrgLicense
        """
        if self._license is None:
            url = url = f"{self._url}/license"
            self._license = KubeOrgLicense(url, self._gis)
        return self._license

    # ----------------------------------------------------------------------
    @property
    def federation(self) -> "KubeOrgFederations":
        """
        Returns manager to work with server federation.

        :return: KubeOrgFederations
        """
        if self._federation is None:
            url = self._url + "/federation"
            self._federation = KubeOrgFederations(url, self._gis)
        return self._federation


###########################################################################
class KubeOrgFederations:
    """
    Provides access to the federation of ArcGIS Server and the ability to
    federate them with the organization.
    """

    _con = None
    _gis = None
    _url = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS") -> "KubeOrgFederations":
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        returns the properties for the Kubernetes License Organization

        :return: dict
        """
        if self._properties is None:
            self._properties = self._con.get(self._url, {"f": "json"})
        return self._properties

    @property
    def servers(self):
        """
        This resource returns detailed information about the ArcGIS Servers
        federated with ArcGIS on Kubernetes. Information such as the ID and
        name of the server, ArcGIS Web Adaptor URL, administration URL, and
        role of the server.
        """
        url = f"{self._url}/servers"
        params = {"f": "json"}
        return self._con.get(path=url, params=params)

    def unfederate(self, server_id: str) -> bool:
        """
        This operation unfederates a currently federated ArcGIS Server from
        your organization. Before performing this operation, the federated
        server should be taken out of read-only mode if it was already in
        that state. This operation is not applicable to the hosting server
        configured as part of the base deployment of ArcGIS Enterprise on
        Kubernetes.

        :returns: Bool
        """
        url = f"{self._url}/servers/{server_id}/unfederate"
        params = {"f": "json"}
        return self._con.post(url, params).pop("success", False)

    def federate(self, url: str, admin_url: str, username: str, password: str) -> bool:
        """
        This operation federates either a GIS Server or ArcGIS Image Server
        with an organization. The federate operation performs a validation
        check to determine whether the provided service and
        dministrative URLs are accessible. If the resulting validation check
        fails, a warning is returned. A SEVERE log type is also returned in
        the organization's logs. After federation, administrators will be
        unable to set a server role for the federated server.

        Once a server has been federated with an organization, services
        that exist on the ArcGIS Server site at the time of federation are
        automatically added to the portal as items. The administrator who
        performs this operation will be assigned as the imported service's
        owner and, once the operation is complete, can reassign ownership
        to other members in the organization. Any subsequent items
        published to the federated server are automatically added as items
        on the portal and are owned by the user who publishes them.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        url                             Required string. The URL of the GIS or image server used by external
                                        users when accessing the server site. If you've added the server to
                                        your organization's reverse proxy server, the URL is the reverse
                                        proxy server address.
        ---------------------------     --------------------------------------------------------------------
        admin_url                       Required string. The URL used to access the server when performing
                                        administrative operations on the internal network. The URL must be
                                        able to be used by the organization to communicate with all servers
                                        in the site, even when one of them is unavailable.
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The username of the primary administrator account
                                        for the server. If this account is disabled, you'll need to
                                        reenable it.
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password of the primary administrator account
                                        for the server.
        ===========================     ====================================================================

        :returns: bool

        """
        params = {
            "f": "json",
            "url": url,
            "adminUrl": admin_url,
            "username": username,
            "password": password,
        }
        url = f"{self._url}/servers/federate"
        return self._con.post(url, params).get("status", "failed") == "success"

    def validate(self, server_id: str) -> dict:
        """
        The validate operation performs validation checks against all
        federated GIS Server and ArcGIS Image Server types within your
        organization, including the hosting server that is built in with an
        ArcGIS Enterprise on Kubernetes deployment. On completion, this
        operation returns status and accessibility information for all
        organization servers. This response also includes any failure
        messages from failed validation checks.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        server_id                       Optional String. When present the validation will occur on that
                                        single server.  If no `server_id` is given, then all servers are
                                        validated.
        ===========================     ====================================================================

        :returns: dict
        """
        if server_id:
            url = f"{self._url}/servers/{server_id}/validate"
        else:
            url = f"{self._url}/servers/validate"
        params = {"f": "json"}

        return self._con.get(url, params)


###########################################################################
class KubeOrgLicense:
    """
    The Licenses resource returns high-level licensing details, such as the
    total number of registered members that can be added, the current
    number of members in the organization, the Enterprise portal version,
    and license manager information. This API endpoint also provides access
    to various operations that allow you to manage your portal licenses for
    your organization.

    """

    _con = None
    _gis = None
    _url = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: "GIS") -> "KubeOrgLicense":
        """
        initializer
        """
        self._url = url
        self._gis = gis
        self._con = gis._con
        self._properties = None

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self) -> dict:
        """
        returns the properties for the Kubernetes License Organization

        :return: dict
        """
        if self._properties is None:
            self._properties = self._con.get(self._url, {"f": "json"})
        return self._properties

    # ----------------------------------------------------------------------
    def export_gdb_license(self, out_folder: str = None) -> str:
        """
        The operation downloads a geodatabaseLicense.ecp file that
        represents the authorization file needed when enabling, creating,
        and updating an enterprise geodatabase in ArcGIS Pro for ArcGIS
        Enterprise on Kubernetes deployments. Accessing this operation
        automatically downloads the .ecp file; no parameters are required
        and no JSON Response is returned for this operation.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        out_folder                      Optional string. The folder where the license file will be saved.
        ===========================     ====================================================================

        :returns: str
        """
        url = self._url + "/exportGeodatabaseLicense"
        return self._con.get(url, params={}, try_json=False, out_folder=out_folder)

    # ----------------------------------------------------------------------
    def update_license_manager(self, config: dict) -> bool:
        """
        This operation allows you to change the license server connection
        information for your portal, as well as register a backup license
        manager for high availability. After changing the license manager
        properties, Portal for ArcGIS automatically restarts to register
        changes and set up connections with the backup license manager.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        config                          Required Dict. The JSON representation of the license server
                                        connection information.

                                        Example: ```{"hostname": "licensemanager.domain.com,backuplicensemanager.domain.com","port": 27000}```
        ===========================     ====================================================================

        :return: Boolean

        """

        url = self._url + "/updateLicenseManager"
        params = {"f": "json", "licenseManagerInfo": json.dumps(config)}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def import_license(self, license_file: str):
        """
        Applies a new license file to a specific organization, which contains the portal's user type and add-on licenses.

        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        license_file                    Required String. The kubernetes license file. For deployments using
                                        ArcGIS Enterprise on Kubernetes 10.9.1 or earlier, this file is an
                                        ArcGIS Enterprise portal license file. For deployments using ArcGIS
                                        Enterprise on Kubernetes 11.0 or later, this is an ArcGIS Enterprise
                                        on Kubernetes license file.
        ===========================     ====================================================================

        :return: Boolean

        """
        params = {"f": "json"}
        url = self._url + "/importLicense"
        file = {"file": license_file}
        res = self._con.post(url, params, files=file)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def validate(self, file, list_ut=False):
        """
        The `validate` operation is used to validate an input license file.
        Only valid license files can be imported into the Enterprise
        portal. If the provided file is valid, the operation will return
        user type, app bundle, and app information from the license file.
        If the file is invalid, the operation will fail and return an error
        message.


        ===========================     ====================================================================
        **Parameter**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        file                            Required String. The kubernetes license file. For deployments using
                                        ArcGIS Enterprise on Kubernetes 10.9.1 or earlier, this file is an
                                        ArcGIS Enterprise portal license file. For deployments using ArcGIS
                                        Enterprise on Kubernetes 11.0 or later, this is an ArcGIS Enterprise
                                        on Kubernetes license file.
        ---------------------------     --------------------------------------------------------------------
        list_ut                         Optional Boolean. Returns a list of user types that are compatible
                                        with the Administrator role. This identifies the user type(s) that
                                        can be assigned to the Initial Administrator Account when creating
                                        a portal.
        ===========================     ====================================================================

        :return: Dict

        """
        file = {"file": file}
        params = {"f": "json", "listAdministratorUserTypes": list_ut}
        url = "%s/validateLicense" % self._url
        res = self._con.post(url, params, files=file)
        return res


###########################################################################
class KubeOrganizations:
    """
    Allows for the management of organizations within the ArcGIS Enterprise
    on Kubernetes deployment.
    """

    _con = None
    _gis = None
    _url = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(
        self, url: str, gis: "GIS", initialize: bool = True
    ) -> "KubeOrganizations":
        """
        Kubernetes Organization
        """
        self._url = url
        self._gis = gis
        self._con = gis._con

        if initialize:
            self._init(gis)

    # ----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""

        params = {"f": "json"}
        try:
            result = self._con.get(self._url, params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< %s @ %s >" % (type(self).__name__, self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url

    # ----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()

    # ----------------------------------------------------------------------
    @property
    def orgs(self) -> tuple:
        """
        Returns a list of registered organizations with the Kubernetes deployment

        :return: tuple
        """
        return tuple(
            [
                KubeOrganization(url=f"{self._url}/{org}", gis=self._gis)
                for org in self.properties["organizations"]
            ]
        )



from __future__ import division

from arcgis.learn._utils.common import ArcGISMSImage

try:
    import random
    import logging
    import math
    import sys
    from collections import namedtuple
    from IPython.display import clear_output

    pyv = sys.version[0]

    from os.path import join, isdir
    from os import mkdir, makedirs
    import os
    import json
    import traceback
    import cv2
    import numpy as np
    import matplotlib.pyplot as plt
    from torch.utils.data import Dataset
    from torch.utils.data import DataLoader
    from .pointcloud_data import get_device
    from fastai.data_block import DataBunch
    import types
    from PIL import Image
    from .env import (
        raise_fastai_import_error,
        HAS_GDAL,
        gdal_import_exception,
        GDAL_INSTALL_MESSAGE,
    )
except Exception:
    import_exception = traceback.format_exc()
    pass

sample_random = random.Random()
sample_random.seed(123456)

Corner = namedtuple("Corner", "x1 y1 x2 y2")
BBox = Corner
Center = namedtuple("Center", "x y w h")
image_name_len = 0

sep = os.sep


def get_image_for_tracking(path, grayscale=False):
    path = str(os.path.abspath(path))
    if not os.path.exists:
        raise Exception(
            f"The image path {path} could not be found on disk, please verify your training data."
        )

    img = ArcGISMSImage.read_image(path, keep_raw=True)
    if img is None:
        return None

    if len(img.shape) >= 3:
        dim = img.shape[0]
        if dim == 2:
            img = np.dstack((img[1], img[0]))[0]
        elif dim > 2:
            img = np.dstack((img[2], img[1], img[0]))

    if grayscale is True and len(img.shape) == 3:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    return img


def corner2center(corner):
    """
    :param corner: Corner or np.array 4*N
    :return: Center or 4 np.array N
    """
    if isinstance(corner, Corner):
        x1, y1, x2, y2 = corner
        return Center((x1 + x2) * 0.5, (y1 + y2) * 0.5, (x2 - x1), (y2 - y1))
    else:
        x1, y1, x2, y2 = corner[0], corner[1], corner[2], corner[3]
        x = (x1 + x2) * 0.5
        y = (y1 + y2) * 0.5
        w = x2 - x1
        h = y2 - y1
        return x, y, w, h


def center2corner(center):
    """
    :param center: Center or np.array 4*N
    :return: Corner or np.array 4*N
    """
    if isinstance(center, Center):
        x, y, w, h = center
        return Corner(x - w * 0.5, y - h * 0.5, x + w * 0.5, y + h * 0.5)
    else:
        x, y, w, h = center[0], center[1], center[2], center[3]
        x1 = x - w * 0.5
        y1 = y - h * 0.5
        x2 = x + w * 0.5
        y2 = y + h * 0.5
        return x1, y1, x2, y2


def cxy_wh_2_rect(pos, sz):
    return np.array([pos[0] - sz[0] / 2, pos[1] - sz[1] / 2, sz[0], sz[1]])  # 0-index


def get_axis_aligned_bbox(region):
    nv = region.size
    if nv == 8:
        cx = np.mean(region[0::2])
        cy = np.mean(region[1::2])
        x1 = min(region[0::2])
        x2 = max(region[0::2])
        y1 = min(region[1::2])
        y2 = max(region[1::2])
        A1 = np.linalg.norm(region[0:2] - region[2:4]) * np.linalg.norm(
            region[2:4] - region[4:6]
        )
        A2 = (x2 - x1) * (y2 - y1)
        s = np.sqrt(A1 / A2)
        w = s * (x2 - x1) + 1
        h = s * (y2 - y1) + 1
    else:
        x = region[0]
        y = region[1]
        w = region[2]
        h = region[3]
        cx = x + w / 2
        cy = y + h / 2

    return cx, cy, w, h


def aug_apply(bbox, param, shape, inv=False, rd=False):
    """
    apply augmentation
    :param bbox: original bbox in image
    :param param: augmentation param, shift/scale
    :param shape: image shape, h, w, (c)
    :param inv: inverse
    :param rd: round bbox
    :return: bbox(, param)
        bbox: augmented bbox
        param: real augmentation param
    """
    if not inv:
        center = corner2center(bbox)
        original_center = center

        real_param = {}
        if "scale" in param:
            scale_x, scale_y = param["scale"]
            imh, imw = shape[:2]
            h, w = center.h, center.w

            scale_x = min(scale_x, float(imw) / w)
            scale_y = min(scale_y, float(imh) / h)
            center = Center(center.x, center.y, center.w * scale_x, center.h * scale_y)

        bbox = center2corner(center)

        if "shift" in param:
            tx, ty = param["shift"]
            x1, y1, x2, y2 = bbox
            imh, imw = shape[:2]

            tx = max(-x1, min(imw - 1 - x2, tx))
            ty = max(-y1, min(imh - 1 - y2, ty))

            bbox = Corner(x1 + tx, y1 + ty, x2 + tx, y2 + ty)

        if rd:
            bbox = Corner(*map(round, bbox))

        current_center = corner2center(bbox)

        real_param["scale"] = (
            current_center.w / original_center.w,
            current_center.h / original_center.h,
        )
        real_param["shift"] = (
            current_center.x - original_center.x,
            current_center.y - original_center.y,
        )

        return bbox, real_param
    else:
        if "scale" in param:
            scale_x, scale_y = param["scale"]
        else:
            scale_x, scale_y = 1.0, 1.0

        if "shift" in param:
            tx, ty = param["shift"]
        else:
            tx, ty = 0, 0

        center = corner2center(bbox)

        center = Center(
            center.x - tx, center.y - ty, center.w / scale_x, center.h / scale_y
        )

        return center2corner(center)


def IoU(rect1, rect2):
    # overlap
    x1, y1, x2, y2 = rect1[0], rect1[1], rect1[2], rect1[3]
    tx1, ty1, tx2, ty2 = rect2[0], rect2[1], rect2[2], rect2[3]

    xx1 = np.maximum(tx1, x1)
    yy1 = np.maximum(ty1, y1)
    xx2 = np.minimum(tx2, x2)
    yy2 = np.minimum(ty2, y2)

    ww = np.maximum(0, xx2 - xx1)
    hh = np.maximum(0, yy2 - yy1)

    area = (x2 - x1) * (y2 - y1)

    target_a = (tx2 - tx1) * (ty2 - ty1)

    inter = ww * hh
    overlap = inter / (area + target_a - inter)

    return overlap


class Anchors:
    def __init__(self, cfg):
        self.stride = 8
        self.ratios = [0.33, 0.5, 1, 2, 3]
        self.scales = [8]
        self.round_dight = 0
        self.image_center = 0
        self.size = 0
        self.anchor_density = 1

        self.__dict__.update(cfg)

        self.anchor_num = (
            len(self.scales) * len(self.ratios) * (self.anchor_density**2)
        )
        self.anchors = None
        self.all_anchors = None
        self.generate_anchors()

    def generate_anchors(self):
        self.anchors = np.zeros((self.anchor_num, 4), dtype=np.float32)

        size = self.stride * self.stride
        count = 0
        anchors_offset = self.stride / self.anchor_density
        anchors_offset = np.arange(self.anchor_density) * anchors_offset
        anchors_offset = anchors_offset - np.mean(anchors_offset)
        x_offsets, y_offsets = np.meshgrid(anchors_offset, anchors_offset)

        for x_offset, y_offset in zip(x_offsets.flatten(), y_offsets.flatten()):
            for r in self.ratios:
                if self.round_dight > 0:
                    ws = round(math.sqrt(size * 1.0 / r), self.round_dight)
                    hs = round(ws * r, self.round_dight)
                else:
                    ws = int(math.sqrt(size * 1.0 / r))
                    hs = int(ws * r)

                for s in self.scales:
                    w = ws * s
                    h = hs * s
                    self.anchors[count][:] = [
                        -w * 0.5 + x_offset,
                        -h * 0.5 + y_offset,
                        w * 0.5 + x_offset,
                        h * 0.5 + y_offset,
                    ][:]
                    count += 1

    def generate_all_anchors(self, im_c, size):
        if self.image_center == im_c and self.size == size:
            return False
        self.image_center = im_c
        self.size = size

        a0x = im_c - size // 2 * self.stride
        ori = np.array([a0x] * 4, dtype=np.float32)
        zero_anchors = self.anchors + ori

        x1 = zero_anchors[:, 0]
        y1 = zero_anchors[:, 1]
        x2 = zero_anchors[:, 2]
        y2 = zero_anchors[:, 3]

        x1, y1, x2, y2 = map(
            lambda x: x.reshape(self.anchor_num, 1, 1), [x1, y1, x2, y2]
        )
        cx, cy, w, h = corner2center([x1, y1, x2, y2])

        disp_x = np.arange(0, size).reshape(1, 1, -1) * self.stride
        disp_y = np.arange(0, size).reshape(1, -1, 1) * self.stride

        cx = cx + disp_x
        cy = cy + disp_y

        zero = np.zeros((self.anchor_num, size, size), dtype=np.float32)
        cx, cy, w, h = map(lambda x: x + zero, [cx, cy, w, h])
        x1, y1, x2, y2 = center2corner([cx, cy, w, h])

        self.all_anchors = np.stack([x1, y1, x2, y2]), np.stack([cx, cy, w, h])
        return True


class SubDataSet(object):
    def __init__(self, cfg):
        for string in ["root", "anno"]:
            if string not in cfg:
                raise Exception('SubDataSet need "{}"'.format(string))

        self.labels = self.filter_zero(cfg["anno"], cfg)

        def isint(x):
            try:
                int(x)
                return True
            except:
                return False

        to_del = []
        for video in self.labels:
            for track in self.labels[video]:
                frames = self.labels[video][track]
                frames = list(map(int, filter(lambda x: isint(x), frames.keys())))
                frames.sort()
                self.labels[video][track]["frames"] = frames
                if len(frames) <= 0:
                    to_del.append((video, track))

        for video, track in to_del:
            del self.labels[video][track]

        to_del = []
        for video in self.labels:
            if len(self.labels[video]) <= 0:
                to_del.append(video)

        for video in to_del:
            del self.labels[video]

        self.videos = list(self.labels.keys())
        self.root = "/"
        self.start = 0
        self.num = len(self.labels)
        self.num_use = self.num
        self.frame_range = 100
        self.path_format = "{}.{}.{}.jpg"
        self.mask_format = "{}.{}.m.png"

        self.pick = []

        self.__dict__.update(cfg)

        self.has_mask = True

        self.num_use = int(self.num_use)

        self.shuffle()

    def filter_zero(self, anno, cfg):
        name = cfg.get("mark", "")

        out = {}
        tot = 0
        new = 0
        zero = 0

        for video, tracks in anno.items():
            new_tracks = {}
            for trk, frames in tracks.items():
                new_frames = {}
                for frm, bbox in frames.items():
                    tot += 1
                    if len(bbox) == 4:
                        x1, y1, x2, y2 = bbox
                        w, h = x2 - x1, y2 - y1
                    else:
                        w, h = bbox
                    if w == 0 or h == 0:
                        zero += 1
                        continue
                    new += 1
                    new_frames[frm] = bbox

                if len(new_frames) > 0:
                    new_tracks[trk] = new_frames

            if len(new_tracks) > 0:
                out[video] = new_tracks

        return out

    def shuffle(self):
        lists = list(range(self.start, self.start + self.num))

        m = 0
        pick = []
        while m < self.num_use:
            sample_random.shuffle(lists)
            pick += lists
            m += self.num

        self.pick = pick[: self.num_use]
        return self.pick

    def get_image_anno(self, video, track, frame):
        frame = "{:06d}".format(frame)
        image_path = join(self.root, video, self.path_format.format(frame, track, "x"))
        image_anno = self.labels[video][track][frame]

        mask_path = join(self.root, video, self.mask_format.format(frame, track))

        return image_path, image_anno, mask_path

    def get_positive_pair(self, index):
        video_name = self.videos[index]
        video = self.labels[video_name]
        track = random.choice(list(video.keys()))
        track_info = video[track]

        frames = track_info["frames"]

        if "hard" not in track_info:
            template_frame = random.randint(0, len(frames) - 1)

            left = max(template_frame - self.frame_range, 0)
            right = min(template_frame + self.frame_range, len(frames) - 1) + 1
            search_range = frames[left:right]
            template_frame = frames[template_frame]
            search_frame = random.choice(search_range)
        else:
            search_frame = random.choice(track_info["hard"])
            left = max(search_frame - self.frame_range, 0)
            right = min(search_frame + self.frame_range, len(frames) - 1) + 1
            template_range = frames[left:right]
            template_frame = random.choice(template_range)
            search_frame = frames[search_frame]

        return (
            self.get_image_anno(video_name, track, template_frame),
            self.get_image_anno(video_name, track, search_frame),
        )

    def get_random_target(self, index=-1):
        if index == -1:
            index = random.randint(0, self.num - 1)
        video_name = self.videos[index]
        video = self.labels[video_name]
        track = random.choice(list(video.keys()))
        track_info = video[track]

        frames = track_info["frames"]
        frame = random.choice(frames)

        return self.get_image_anno(video_name, track, frame)


def crop_hwc(image, bbox, out_sz, padding=(0, 0, 0)):
    bbox = [float(x) for x in bbox]
    a = (out_sz - 1) / (bbox[2] - bbox[0])
    b = (out_sz - 1) / (bbox[3] - bbox[1])
    c = -a * bbox[0]
    d = -b * bbox[1]
    mapping = np.array([[a, 0, c], [0, b, d]]).astype(float)
    crop = cv2.warpAffine(
        image,
        mapping,
        (out_sz, out_sz),
        borderMode=cv2.BORDER_CONSTANT,
        borderValue=padding,
    )
    return crop


class Augmentation:
    def __init__(self, cfg):
        self.shift = 0
        self.scale = 0
        self.blur = 0  # False
        self.resize = False
        self.rgbVar = np.array(
            [
                [-0.55919361, 0.98062831, -0.41940627],
                [1.72091413, 0.19879334, -1.82968581],
                [4.64467907, 4.73710203, 4.88324118],
            ],
            dtype=np.float32,
        )
        self.flip = 0

        self.eig_vec = np.array(
            [
                [0.4009, 0.7192, -0.5675],
                [-0.8140, -0.0045, -0.5808],
                [0.4203, -0.6948, -0.5836],
            ],
            dtype=np.float32,
        )

        self.eig_val = np.array([[0.2175, 0.0188, 0.0045]], np.float32)

        self.__dict__.update(cfg)

    @staticmethod
    def random():
        return random.random() * 2 - 1.0

    def blur_image(self, image):
        def rand_kernel():
            size = np.random.randn(1)
            size = int(np.round(size)) * 2 + 1
            if size < 0:
                return None
            if random.random() < 0.5:
                return None
            size = min(size, 45)
            kernel = np.zeros((size, size))
            c = int(size / 2)
            wx = random.random()
            kernel[:, c] += 1.0 / size * wx
            kernel[c, :] += 1.0 / size * (1 - wx)
            return kernel

        kernel = rand_kernel()

        if kernel is not None:
            image = cv2.filter2D(image, -1, kernel)
        return image

    def __call__(self, image, bbox, size, gray=False, mask=None):
        if gray:
            grayed = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            image = np.zeros((grayed.shape[0], grayed.shape[1], 3), np.uint8)
            image[:, :, 0] = image[:, :, 1] = image[:, :, 2] = grayed

        shape = image.shape

        crop_bbox = center2corner((shape[0] // 2, shape[1] // 2, size - 1, size - 1))

        param = {}
        if self.shift:
            param["shift"] = (
                Augmentation.random() * self.shift,
                Augmentation.random() * self.shift,
            )

        if self.scale:
            param["scale"] = (
                (1.0 + Augmentation.random() * self.scale),
                (1.0 + Augmentation.random() * self.scale),
            )

        crop_bbox, _ = aug_apply(Corner(*crop_bbox), param, shape)

        x1 = crop_bbox.x1
        y1 = crop_bbox.y1

        bbox = BBox(bbox.x1 - x1, bbox.y1 - y1, bbox.x2 - x1, bbox.y2 - y1)

        if self.scale:
            scale_x, scale_y = param["scale"]
            bbox = Corner(
                bbox.x1 / scale_x,
                bbox.y1 / scale_y,
                bbox.x2 / scale_x,
                bbox.y2 / scale_y,
            )

        image = crop_hwc(image, crop_bbox, size)
        if not mask is None:
            mask = crop_hwc(mask, crop_bbox, size)

        offset = np.dot(self.rgbVar, np.random.randn(3, 1))
        offset = offset[::-1]
        offset = offset.reshape(3)
        image = image - offset

        if self.blur > random.random():
            image = self.blur_image(image)

        if self.resize:
            imageSize = image.shape[:2]
            ratio = max(math.pow(random.random(), 0.5), 0.2)  # 25 ~ 255
            rand_size = (
                int(round(ratio * imageSize[0])),
                int(round(ratio * imageSize[1])),
            )
            image = cv2.resize(image, rand_size)
            image = cv2.resize(image, tuple(imageSize))

        if self.flip and self.flip > Augmentation.random():
            image = cv2.flip(image, 1)
            mask = cv2.flip(mask, 1)
            width = image.shape[1]
            bbox = Corner(width - 1 - bbox.x2, bbox.y1, width - 1 - bbox.x1, bbox.y2)

        return image, bbox, mask


class AnchorTargetLayer:
    def __init__(self, cfg):
        self.thr_high = 0.6
        self.thr_low = 0.3
        self.negative = 16
        self.rpn_batch = 64
        self.positive = 16

        self.__dict__.update(cfg)

    def __call__(self, anchor, target, size, neg=False, need_iou=False):
        anchor_num = anchor.anchors.shape[0]

        cls = np.zeros((anchor_num, size, size), dtype=np.int64)
        cls[...] = -1
        delta = np.zeros((4, anchor_num, size, size), dtype=np.float32)
        delta_weight = np.zeros((anchor_num, size, size), dtype=np.float32)

        def select(position, keep_num=16):
            num = position[0].shape[0]
            if num <= keep_num:
                return position, num
            slt = np.arange(num)
            np.random.shuffle(slt)
            slt = slt[:keep_num]
            return tuple(p[slt] for p in position), keep_num

        if neg:
            l = size // 2 - 3
            r = size // 2 + 3 + 1

            cls[:, l:r, l:r] = 0

            neg, neg_num = select(np.where(cls == 0), self.negative)
            cls[:] = -1
            cls[neg] = 0

            if not need_iou:
                return cls, delta, delta_weight
            else:
                overlap = np.zeros((anchor_num, size, size), dtype=np.float32)
                return cls, delta, delta_weight, overlap

        tcx, tcy, tw, th = corner2center(target)

        anchor_box = anchor.all_anchors[0]
        anchor_center = anchor.all_anchors[1]
        x1, y1, x2, y2 = anchor_box[0], anchor_box[1], anchor_box[2], anchor_box[3]
        cx, cy, w, h = (
            anchor_center[0],
            anchor_center[1],
            anchor_center[2],
            anchor_center[3],
        )

        delta[0] = (tcx - cx) / w
        delta[1] = (tcy - cy) / h
        delta[2] = np.log(tw / w)
        delta[3] = np.log(th / h)

        overlap = IoU([x1, y1, x2, y2], target)

        pos = np.where(overlap > self.thr_high)
        neg = np.where(overlap < self.thr_low)

        pos, pos_num = select(pos, self.positive)
        neg, neg_num = select(neg, self.rpn_batch - pos_num)

        cls[pos] = 1
        delta_weight[pos] = 1.0 / (pos_num + 1e-6)

        cls[neg] = 0

        if not need_iou:
            return cls, delta, delta_weight
        else:
            return cls, delta, delta_weight, overlap


class DataSets(Dataset):
    def __init__(
        self,
        data_path,
        annotation,
        num,
        num_epoch=1,
        image_name_len=6,
        annotations=None,
    ):
        super(DataSets, self).__init__()

        anchor = {
            "stride": 8,
            "ratios": [0.33, 0.5, 1, 2, 3],
            "scales": [8],
            "round_dight": 0,
        }
        self.anchors = Anchors(anchor)

        self.template_size = 127
        self.origin_size = 127
        self.search_size = 143
        self.size = 3
        self.base_size = 0
        self.image_name_len = image_name_len
        self.annotations = annotations

        self.crop_size = 0
        self.chip_size = 127
        self._is_multispectral = False

        if (
            self.search_size - self.template_size
        ) / self.anchors.stride + 1 + self.base_size != self.size:
            raise Exception("size not match!")

        self.template_small = False

        self.anchors.generate_all_anchors(im_c=self.search_size // 2, size=self.size)

        self.anchor_target = AnchorTargetLayer({})

        datasets = {
            "siammask_data": {
                "root": os.path.join(data_path, "crop"),
                "anno": annotation,
                "num_use": 200000,
                "frame_range": 20,
            }
        }

        self.all_data = []
        start = 0
        self.num = 0
        self.show_batch_data = annotation
        for name in datasets:
            dataset = datasets[name]
            dataset["mark"] = name
            dataset["start"] = start

            dataset = SubDataSet(dataset)
            self.all_data.append(dataset)

            start += dataset.num
            self.num += dataset.num_use

        aug_cfg = {
            "template": {"shift": 4, "scale": 0.05},
            "search": {"shift": 64, "scale": 0.18, "blur": 0.18},
            "neg": 0.2,
            "gray": 0.25,
        }
        self.template_aug = Augmentation(aug_cfg["template"])
        self.search_aug = Augmentation(aug_cfg["search"])
        self.gray = aug_cfg["gray"]
        self.neg = aug_cfg["neg"]
        self.inner_neg = 0 if "inner_neg" not in aug_cfg else aug_cfg["inner_neg"]

        self.pick = None
        self.num = int(num)
        self.num *= num_epoch
        self.shuffle()

        self.infos = {
            "template": self.template_size,
            "search": self.search_size,
            "template_small": self.template_small,
            "gray": self.gray,
            "neg": self.neg,
            "inner_neg": self.inner_neg,
            "crop_size": self.crop_size,
            "anchor_target": self.anchor_target.__dict__,
            "num": self.num // num_epoch,
        }
        self._img_exts = [".jpg", ".png", ".tif"]

    def imread(self, path):
        img = get_image_for_tracking(path)

        if self.origin_size == self.template_size:
            return img, 1.0

        def map_size(exe, size):
            return int(round(((exe + 1) / (self.origin_size + 1) * (size + 1) - 1)))

        nsize = map_size(self.template_size, img.shape[1])

        img = cv2.resize(img, (nsize, nsize))

        return img, nsize / img.shape[1]

    def shuffle(self):
        pick = []
        m = 0
        while m < self.num:
            p = []
            for subset in self.all_data:
                sub_p = subset.shuffle()
                p += sub_p

            sample_random.shuffle(p)

            pick += p
            m = len(pick)
        self.pick = pick

    def __len__(self):
        return self.num

    def find_dataset(self, index):
        for dataset in self.all_data:
            if dataset.start + dataset.num > index:
                return dataset, index - dataset.start

    def __getitem__(self, index, debug=False, show_batch=False, show_results=False):
        index = self.pick[index]
        dataset, index = self.find_dataset(index)

        gray = self.gray and self.gray > random.random()
        neg = self.neg and self.neg > random.random()

        if neg:
            template = dataset.get_random_target(index)
            if self.inner_neg and self.inner_neg > random.random():
                search = dataset.get_random_target()
            else:
                search = random.choice(self.all_data).get_random_target()
        else:
            template, search = dataset.get_positive_pair(index)

        display_image = search

        def center_crop(img, size):
            shape = img.shape[1]
            if shape == size:
                return img
            c = shape // 2
            l = c - size // 2
            r = c + size // 2 + 1
            return img[l:r, l:r]

        template_image, scale_z = self.imread(template[0])

        if self.template_small:
            template_image = center_crop(template_image, self.template_size)

        search_image, scale_x = self.imread(search[0])
        if dataset.has_mask and not neg:
            search_mask = (
                get_image_for_tracking(search[2], grayscale=True) > 0
            ).astype(np.float32)
        else:
            search_mask = np.zeros(search_image.shape[:2], dtype=np.float32)

        if self.crop_size > 0:
            search_image = center_crop(search_image, self.crop_size)
            search_mask = center_crop(search_mask, self.crop_size)

        def toBBox(image, shape):
            imh, imw = image.shape[:2]
            if len(shape) == 4:
                w, h = shape[2] - shape[0], shape[3] - shape[1]
            else:
                w, h = shape
            context_amount = 0.5
            exemplar_size = self.template_size
            wc_z = w + context_amount * (w + h)
            hc_z = h + context_amount * (w + h)
            s_z = np.sqrt(wc_z * hc_z)
            scale_z = exemplar_size / s_z
            w = w * scale_z
            h = h * scale_z
            cx, cy = imw // 2, imh // 2
            bbox = center2corner(Center(cx, cy, w, h))
            return bbox

        template_box = toBBox(template_image, template[1])
        search_box = toBBox(search_image, search[1])

        template, _, _ = self.template_aug(
            template_image, template_box, self.template_size, gray=gray
        )
        search, bbox, mask = self.search_aug(
            search_image, search_box, self.search_size, gray=gray, mask=search_mask
        )

        if show_batch:
            return display_image, self.image_name_len, self.annotations

        if show_results:
            return display_image

        cls, delta, delta_weight = self.anchor_target(
            self.anchors, bbox, self.size, neg
        )
        if dataset.has_mask and not neg:
            mask_weight = cls.max(axis=0, keepdims=True)
        else:
            mask_weight = np.zeros([1, cls.shape[1], cls.shape[2]], dtype=np.float32)

        template, search = map(
            lambda x: np.transpose(x, (2, 0, 1)).astype(np.float32), [template, search]
        )

        mask = (np.expand_dims(mask, axis=0) > 0.5) * 2 - 1

        return (
            [
                template,
                search,
                cls,
                delta,
                delta_weight,
                np.array(bbox, np.float32),
                np.array(mask, np.float32),
                np.array(mask_weight, np.float32),
            ],
            [[]],
        )

    def show(self, idx, axes=None):
        global image_name_len
        if axes is None:
            _, axes = plt.subplots(1, 2, figsize=(15, 7))

        for i, id in enumerate(idx):
            data, img_len, ann = self.__getitem__(id, debug=False, show_batch=True)
            sep = os.sep

            base_folder = data[0].split("crop")[0]
            folder_name = data[0].split(sep)[-2]
            frame_id = data[0].split(sep)[-1].split(".")[1]
            frame_number = data[0].split(sep)[-1].split(".")[0]
            mask_name = folder_name + sep + frame_number.zfill(img_len)
            # key = list(ann[folder_name].keys())[-1]
            for key in list(ann[folder_name].keys()):
                for cnt in ann[folder_name][key]:
                    if cnt["mask_name"] == mask_name:
                        display_data = (cnt["display_mask"], cnt["bbox"])
                        break

            img_path = os.path.join(
                base_folder,
                "JPEGImages",
                folder_name,
                frame_number.zfill(img_len) + ".jpg",
            )

            image = get_image_for_tracking(img_path)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            bbox = display_data[1]
            cv2.rectangle(
                image,
                (int(bbox[0]), int(bbox[1])),
                (int(bbox[2]) - int(bbox[0]), int(bbox[3] - int(bbox[1]))),
                (255, 0, 0),
                2,
            )

            overlay = image.copy()
            cv2.drawContours(overlay, display_data[0], -1, (0, 0, 255), -1)

            image_new = cv2.addWeighted(overlay, 0.5, image, 1, 0)
            axes[i].set_xticks([])
            axes[i].set_yticks([])
            axes[i].imshow(image_new)

    def _show_pro(self, idx, axes=None):
        global image_name_len
        if axes is None:
            _, axes = plt.subplots(1, 2, figsize=(15, 7))

        for i, id in enumerate(idx):
            data, img_len, ann = self.__getitem__(id, debug=False, show_batch=True)
            sep = os.sep

            base_folder = data[0].split("crop")[0]
            folder_name = data[0].split(sep)[-2]
            frame_id = data[0].split(sep)[-1].split(".")[1]
            frame_number = data[0].split(sep)[-1].split(".")[0]
            mask_name = folder_name + sep + frame_number.zfill(img_len)
            for key in list(ann[folder_name].keys()):
                for cnt in ann[folder_name][key]:
                    if cnt["mask_name"] == mask_name:
                        display_data = (cnt["display_mask"], cnt["bbox"])
                        break

            for ext in self._img_exts:
                try:
                    img_path = os.path.join(
                        base_folder,
                        "images",
                        frame_number.zfill(img_len) + ext,
                    )
                    if os.path.isfile(img_path):
                        break
                except:
                    continue

            image = get_image_for_tracking(img_path)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            bbox = display_data[1]
            cv2.rectangle(
                image,
                (int(bbox[0]), int(bbox[1])),
                (int(bbox[2]) - int(bbox[0]), int(bbox[3] - int(bbox[1]))),
                (255, 0, 0),
                2,
            )

            overlay = image.copy()
            cv2.drawContours(overlay, display_data[0], -1, (0, 0, 255), -1)

            image_new = cv2.addWeighted(overlay, 0.5, image, 1, 0)
            axes[i].set_xticks([])
            axes[i].set_yticks([])
            axes[i].imshow(image_new)


def xyxy_to_xywh(xyxy):
    """Convert [x1 y1 x2 y2] box format to [x1 y1 w h] format."""
    if isinstance(xyxy, (list, tuple)):
        assert len(xyxy) == 4
        x1, y1 = xyxy[0], xyxy[1]
        w = xyxy[2] - x1 + 1
        h = xyxy[3] - y1 + 1
        return (x1, y1, w, h)
    elif isinstance(xyxy, np.ndarray):
        return np.hstack((xyxy[:, 0:2], xyxy[:, 2:4] - xyxy[:, 0:2] + 1))
    else:
        raise TypeError("Argument xyxy must be a list, tuple, or numpy array.")


def polys_to_boxes(polys):
    """Convert a list of polygons into an array of tight bounding boxes."""
    boxes_from_polys = np.zeros((len(polys), 4), dtype=np.float32)
    for i in range(len(polys)):
        poly = polys[i]
        x0 = min(min(p[::2]) for p in poly)
        x1 = max(max(p[::2]) for p in poly)
        y0 = min(min(p[1::2]) for p in poly)
        y1 = max(max(p[1::2]) for p in poly)
        boxes_from_polys[i, :] = [x0, y0, x1, y1]
    return boxes_from_polys


def crop_hwc(image, bbox, out_sz, padding=(0, 0, 0)):
    a = (out_sz - 1) / (bbox[2] - bbox[0])
    b = (out_sz - 1) / (bbox[3] - bbox[1])
    c = -a * bbox[0]
    d = -b * bbox[1]
    mapping = np.array([[a, 0, c], [0, b, d]]).astype(float)
    crop = cv2.warpAffine(
        image,
        mapping,
        (out_sz, out_sz),
        borderMode=cv2.BORDER_CONSTANT,
        borderValue=padding,
    )
    return crop


def pos_s_2_bbox(pos, s):
    return [pos[0] - s / 2, pos[1] - s / 2, pos[0] + s / 2, pos[1] + s / 2]


def crop_like_SiamFC(
    image,
    bbox,
    context_amount=0.5,
    exemplar_size=127,
    instanc_size=255,
    padding=(0, 0, 0),
):
    target_pos = [(bbox[2] + bbox[0]) / 2.0, (bbox[3] + bbox[1]) / 2.0]
    target_size = [bbox[2] - bbox[0], bbox[3] - bbox[1]]
    wc_z = target_size[1] + context_amount * sum(target_size)
    hc_z = target_size[0] + context_amount * sum(target_size)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = exemplar_size / s_z
    d_search = (instanc_size - exemplar_size) / 2
    pad = d_search / scale_z
    s_x = s_z + 2 * pad

    z = crop_hwc(image, pos_s_2_bbox(target_pos, s_z), exemplar_size, padding)
    x = crop_hwc(image, pos_s_2_bbox(target_pos, s_x), instanc_size, padding)
    return z, x


def crop_like_SiamFCx(
    image,
    bbox,
    context_amount=0.5,
    exemplar_size=127,
    instanc_size=255,
    padding=(0, 0, 0),
):
    target_pos = [(bbox[2] + bbox[0]) / 2.0, (bbox[3] + bbox[1]) / 2.0]
    target_size = [bbox[2] - bbox[0], bbox[3] - bbox[1]]
    wc_z = target_size[1] + context_amount * sum(target_size)
    hc_z = target_size[0] + context_amount * sum(target_size)
    s_z = np.sqrt(wc_z * hc_z)
    scale_z = exemplar_size / s_z
    d_search = (instanc_size - exemplar_size) / 2
    pad = d_search / scale_z
    s_x = s_z + 2 * pad

    x = crop_hwc(image, pos_s_2_bbox(target_pos, s_x), instanc_size, padding)
    return x


def crop_video_pro(video, v, crop_path, data_path, instanc_size):
    video_crop_base_path = join(crop_path, video)
    if not isdir(video_crop_base_path):
        makedirs(video_crop_base_path)

    anno_base_path = join(data_path, "labels")
    img_base_path = join(data_path, "images")

    for trackid, o in enumerate(list(v)):
        obj = v[o]
        for frame in obj:
            file_name = frame["file_name"]
            ann_ext = frame["ext"]
            ann_path = join(join(anno_base_path, video), file_name + ann_ext)
            img_path = join(img_base_path, file_name + ann_ext)

            if not os.path.isfile(img_path) and ann_ext == ".png":
                img_path = join(img_base_path, file_name + ".jpg")

            label = get_image_for_tracking(ann_path, grayscale=True)

            im = get_image_for_tracking(img_path)
            if im is None:
                continue

            avg_chans = np.mean(im, axis=(0, 1))
            bbox = frame["bbox"]
            bbox[2] += bbox[0]
            bbox[3] += bbox[1]
            x = crop_like_SiamFCx(
                im, bbox, instanc_size=instanc_size, padding=avg_chans
            )

            cv2.imwrite(
                join(
                    video_crop_base_path,
                    "{:06d}.{:02d}.x.jpg".format(
                        int(file_name.split(sep)[-1]), trackid
                    ),
                ),
                x,
            )

            mask = crop_like_SiamFCx(
                (label == int(o)).astype(np.float32),
                bbox,
                instanc_size=instanc_size,
                padding=0,
            )
            mask = ((mask > 0.2) * 255).astype(np.uint8)
            x[:, :, 0] = mask + (mask == 0) * x[:, :, 0]
            cv2.imwrite(
                join(
                    video_crop_base_path,
                    "{:06d}.{:02d}.m.png".format(
                        int(file_name.split(sep)[-1]), trackid
                    ),
                ),
                mask,
            )


def crop_video(video, v, crop_path, data_path, instanc_size):
    video_crop_base_path = join(crop_path, video)
    if not isdir(video_crop_base_path):
        makedirs(video_crop_base_path)

    anno_base_path = join(data_path, "Annotations")
    img_base_path = join(data_path, "JPEGImages")

    for trackid, o in enumerate(list(v)):
        obj = v[o]
        for frame in obj:
            file_name = frame["file_name"]
            ann_path = join(anno_base_path, file_name + ".png")
            img_path = join(img_base_path, file_name + ".jpg")
            im = get_image_for_tracking(img_path)
            label = get_image_for_tracking(ann_path, grayscale=True)
            avg_chans = np.mean(im, axis=(0, 1))
            bbox = frame["bbox"]
            bbox[2] += bbox[0]
            bbox[3] += bbox[1]
            x = crop_like_SiamFCx(
                im, bbox, instanc_size=instanc_size, padding=avg_chans
            )
            cv2.imwrite(
                join(
                    video_crop_base_path,
                    "{:06d}.{:02d}.x.jpg".format(
                        int(file_name.split(sep)[-1]), trackid
                    ),
                ),
                x,
            )
            mask = crop_like_SiamFCx(
                (label == int(o)).astype(np.float32),
                bbox,
                instanc_size=instanc_size,
                padding=0,
            )
            mask = ((mask > 0.2) * 255).astype(np.uint8)
            x[:, :, 0] = mask + (mask == 0) * x[:, :, 0]
            cv2.imwrite(
                join(
                    video_crop_base_path,
                    "{:06d}.{:02d}.m.png".format(
                        int(file_name.split(sep)[-1]), trackid
                    ),
                ),
                mask,
            )


class Instance(object):
    instID = 0
    pixelCount = 0

    def __init__(self, imgNp, instID):
        if instID == 0:
            return
        self.instID = int(instID)
        self.pixelCount = int(self.getInstancePixels(imgNp, instID))

    def getInstancePixels(self, imgNp, instLabel):
        return (imgNp == instLabel).sum()

    def toDict(self):
        buildDict = {}
        buildDict["instID"] = self.instID
        buildDict["pixelCount"] = self.pixelCount
        return buildDict

    def __str__(self):
        return "(" + str(self.instID) + ")"


def train_val_split(dataset):
    snippets = dict()
    for k, v in dataset.items():
        video = dict()
        for i, o in enumerate(list(v)):
            obj = v[o]
            snippet = dict()
            trackid = "{:02d}".format(i)
            for frame in obj:
                file_name = frame["file_name"]
                frame_name = "{:06d}".format(int(file_name.split(sep)[-1]))
                bbox = frame["bbox"]
                bbox[2] += bbox[0]
                bbox[3] += bbox[1]
                snippet[frame_name] = bbox
            video[trackid] = snippet
        snippets[k] = video
    return snippets


def printProgressBar(i, max):
    n_bar = 50
    j = i / max
    sys.stdout.write("\r")
    sys.stdout.write(f"[{'=' * int(n_bar * j):{n_bar}s}] {int(100 * j)}%  completed")
    sys.stdout.flush()


def check_data_sanity(path):
    if not os.path.isdir(path):
        raise Exception(f"Invalid directory. Please check the path {path}")

    ann_dir_path = os.path.join(path, "Annotations")
    if not os.path.isdir(ann_dir_path):
        raise Exception(
            f"Invalid directory. Please check the "
            f"annotation folder path {ann_dir_path}. "
            "Please make sure the folder name is Annotations"
        )

    ann_dirs = os.listdir(ann_dir_path)
    img_dir_path = os.path.join(path, "JPEGImages")
    if not os.path.isdir(img_dir_path):
        raise Exception(
            f"Invalid directory. Please check the "
            f"images folder path {img_dir_path}. "
            "Please make sure the folder name is JPEGImages"
        )

    img_dirs = os.listdir(img_dir_path)
    json_file_path = os.path.join(path, "meta.json")
    if not os.path.isfile(json_file_path):
        raise Exception(
            f"Invalid file. Please check the "
            f"meta.json file path {json_file_path}. "
            "Please make sure the name of json is meta.json"
        )
    json_ann = json.load(open(json_file_path))
    total = 0
    for _, video in enumerate(json_ann["videos"]):
        if video in ann_dirs and video in img_dirs:
            total += 1
    if total <= 1:
        raise Exception(
            f"Please input at least two sequences in the {path.name}"
            " directory namely 'Annotations' and 'JPEGImages'."
            "Also, please provide at least two sequences "
            "in meta.json"
        )


def check_pro_data_sanity(path):
    if not os.path.isdir(path):
        raise Exception(f"Invalid directory. Please check the path {path}")

    ann_dir_path = os.path.join(path, "labels")
    if not os.path.isdir(ann_dir_path):
        raise Exception(
            f"Invalid directory. Please check the "
            f"annotation folder path {ann_dir_path}. "
            "Please make sure the folder name is labels"
        )

    img_dir_path = os.path.join(path, "images")
    if not os.path.isdir(img_dir_path):
        raise Exception(
            f"Invalid directory. Please check the "
            f"images folder path {img_dir_path}. "
            "Please make sure the folder name is images"
        )

    total = 0
    seq_paths = [
        f.path
        for f in os.scandir(ann_dir_path)
        if f.is_dir() and len(os.listdir(f.path)) != 0
    ]
    for idx, seq_path in enumerate(seq_paths):
        annotations = [
            file
            for file in os.listdir(seq_path)
            if file.endswith(".png") or file.endswith(".jpg") or file.endswith(".tif")
        ]
        if len(annotations) > 0:
            total += 1

    if total <= 1:
        raise Exception(
            f"Please input at least two sequences in the {path.name}"
            " directory namely 'labels'."
        )


def prepare_pro_data(path, batch_size, val_split_pct, **kwargs):
    check_pro_data_sanity(path)
    global image_name_len
    data_dir = path
    ann_dir = os.path.join(path, "labels")
    img_dir = os.path.join(path, "images")
    seq_paths = [
        f.path
        for f in os.scandir(ann_dir)
        if f.is_dir() and len(os.listdir(f.path)) != 0
    ]
    num_obj = 0
    num_ann = 0
    all_objects = 0
    print("Extracting info")
    ann_dict = {}
    total = len(seq_paths)
    for idx, seq_path in enumerate(seq_paths):
        seq_name = os.path.basename(seq_path)
        annotation_files = [
            file
            for file in os.listdir(seq_path)
            if file.endswith(".png") or file.endswith(".jpg") or file.endswith(".tif")
        ]
        annotations = []
        instanceIds = []
        for annotation in annotation_files:
            ann_filename = os.path.join(seq_path, annotation)
            file_name_suffix = os.path.basename(ann_filename)
            file_name_suffix_split = os.path.splitext(file_name_suffix)[0]
            file_name_ext = os.path.splitext(file_name_suffix)[1]
            image_name_len = len(file_name_suffix_split)

            img = get_image_for_tracking(ann_filename, grayscale=True)

            if img is None:
                continue

            h, w = img.shape[:2]
            objects = dict()
            for instanceId in np.unique(img):
                if instanceId == 0:
                    continue
                instance_obj = Instance(img, instanceId)
                instance_obj_dict = instance_obj.toDict()
                mask = (img == instanceId).astype(np.uint8)
                contour, _ = cv2.findContours(
                    mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
                )
                polygons = [c.reshape(-1).tolist() for c in contour]
                instance_obj_dict["display_mask"] = contour
                instance_obj_dict["contours"] = [p for p in polygons if len(p) > 4]
                if (
                    len(instance_obj_dict["contours"])
                    and instance_obj_dict["pixelCount"] > 1000
                ):
                    objects[instanceId] = instance_obj_dict

            for objId in objects:
                if len(objects[objId]) == 0:
                    continue
                obj = objects[objId]
                len_p = [len(p) for p in obj["contours"]]
                if min(len_p) <= 4:
                    print("Warning: invalid contours.")
                    continue
                ann = dict()
                ann["h"] = h
                ann["w"] = w
                ann["file_name"] = file_name_suffix_split
                ann["id"] = int(objId)
                ann["segmentation"] = obj["contours"]
                ann["iscrowd"] = 0
                ann["display_mask"] = obj["display_mask"]
                ann["area"] = obj["pixelCount"]
                ann["bbox"] = xyxy_to_xywh(polys_to_boxes([obj["contours"]])).tolist()[
                    0
                ]
                ann["mask_name"] = seq_name + os.sep + file_name_suffix_split
                ann["ext"] = file_name_ext
                annotations.append(ann)
                all_objects += 1
                instanceIds.append(objId)
                num_ann += 1

        instanceIds = sorted(set(instanceIds))
        num_obj += len(instanceIds)
        video_ann = {str(iId): [] for iId in instanceIds}
        for ann in annotations:
            video_ann[str(ann["id"])].append(ann)

        ann_dict[seq_name] = video_ann
        printProgressBar(idx, total)

    items = list(ann_dict.items())
    train_dict = dict(items)

    clear_output()
    crop_path = os.path.join(path, "crop")
    if not os.path.isdir(crop_path):
        os.mkdir(crop_path)
    set_crop_base_path = join(crop_path)
    set_img_base_path = data_dir
    n_video = len(train_dict)

    print("Applying transformations..")

    total = len(train_dict.keys())
    ind_pb = 0
    for k, v in train_dict.items():
        try:
            crop_video_pro(k, v, set_crop_base_path, set_img_base_path, 511)
        except Exception as e:
            print(e)
            break
        ind_pb += 1
        printProgressBar(ind_pb, total)

    val_set = int(((n_video * (val_split_pct * 10)) / 100) * 10)
    val_all_obj = int(((all_objects * (val_split_pct * 10)) / 100) * 10)
    if val_set == 0:
        val_set = 1

    def num_anns(elem):
        ann_dict = elem[1]
        val = 0
        for k, v in ann_dict.items():
            val = val + len(v)
        return val

    items = sorted(items, key=num_anns, reverse=True)
    train_dict = dict(items[:-val_set])
    snippets = train_val_split(train_dict)
    train = {k: v for (k, v) in snippets.items()}
    val_dict = dict(items[-val_set:])

    snippets = train_val_split(val_dict)
    val = {k: v for (k, v) in snippets.items()}

    if all_objects == 0:
        raise Exception(
            "Not enough valid data. Please ensure the labels are correct and enough data is present."
        )

    train_set = DataSets(path, train, all_objects, 5, image_name_len, ann_dict)
    val_set = DataSets(path, val, val_all_obj, 5, image_name_len)

    train_set.shuffle()
    val_set.shuffle()
    init_kwargs = {}
    train_dl = DataLoader(
        train_set,
        batch_size=batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=None,
        **init_kwargs,
    )

    valid_dl = DataLoader(
        val_set,
        batch_size=batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=None,
        **init_kwargs,
    )

    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device)
    data.path = path
    data.infos = data.train_ds.infos
    data.show_batch = types.MethodType(_show_batch_pro, data)
    data._dataset_type = "ObjectTracking"
    data.train_folders = train
    data.val_folders = val
    clear_output()
    return data


def prepare_object_tracking_data(path, batch_size, val_split_pct, **kwargs):
    check_data_sanity(path)
    global image_name_len
    data_dir = path
    ann_dir = os.path.join(path, "Annotations")
    num_obj = 0
    num_ann = 0
    all_objects = 0
    print("Extracting info")
    ann_dict = {}
    json_ann = json.load(open(os.path.join(data_dir, "meta.json")))
    total = len(json_ann["videos"])
    for vid, video in enumerate(json_ann["videos"]):
        if video not in os.listdir(ann_dir):
            continue
        v = json_ann["videos"][video]
        frames = []
        for obj in v["objects"]:
            o = v["objects"][obj]
            frames.extend(o["frames"])
        frames = sorted(set(frames))
        annotations = []
        instanceIds = []
        for frame in frames:
            file_name = os.path.join(video, frame)
            mask_name = video + os.sep + frame
            mask_filename = os.path.join(ann_dir, file_name + ".png")
            image_name_len = len(frame)
            img = get_image_for_tracking(mask_filename, grayscale=True)
            if img is None:
                continue
            h, w = img.shape[:2]
            objects = dict()
            for instanceId in np.unique(img):
                if instanceId == 0:
                    continue
                instance_obj = Instance(img, instanceId)
                instance_obj_dict = instance_obj.toDict()
                mask = (img == instanceId).astype(np.uint8)
                contour, _ = cv2.findContours(
                    mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE
                )
                polygons = [c.reshape(-1).tolist() for c in contour]
                instance_obj_dict["display_mask"] = contour
                instance_obj_dict["contours"] = [p for p in polygons if len(p) > 4]
                if (
                    len(instance_obj_dict["contours"])
                    and instance_obj_dict["pixelCount"] > 1000
                ):
                    objects[instanceId] = instance_obj_dict

            for objId in objects:
                if len(objects[objId]) == 0:
                    continue
                obj = objects[objId]
                len_p = [len(p) for p in obj["contours"]]
                if min(len_p) <= 4:
                    print("Warning: invalid contours.")
                    continue

                ann = dict()
                ann["h"] = h
                ann["w"] = w
                ann["file_name"] = file_name
                ann["id"] = int(objId)
                ann["segmentation"] = obj["contours"]
                ann["iscrowd"] = 0
                ann["display_mask"] = obj["display_mask"]
                ann["area"] = obj["pixelCount"]
                ann["bbox"] = xyxy_to_xywh(polys_to_boxes([obj["contours"]])).tolist()[
                    0
                ]
                ann["mask_name"] = mask_name
                annotations.append(ann)
                all_objects += 1
                instanceIds.append(objId)
                num_ann += 1

        instanceIds = sorted(set(instanceIds))
        num_obj += len(instanceIds)
        video_ann = {str(iId): [] for iId in instanceIds}
        for ann in annotations:
            video_ann[str(ann["id"])].append(ann)

        ann_dict[video] = video_ann
        printProgressBar(vid, total)

    items = list(ann_dict.items())
    train_dict = dict(items)

    clear_output()
    crop_path = os.path.join(path, "crop")
    if not isdir(crop_path):
        mkdir(crop_path)
    set_crop_base_path = join(crop_path)
    set_img_base_path = data_dir
    n_video = len(train_dict)
    print("Applying transformations..")
    total = len(train_dict.keys())
    ind_pb = 0
    for k, v in train_dict.items():
        try:
            crop_video(k, v, set_crop_base_path, set_img_base_path, 511)
        except Exception as e:
            print(e)
            break
        ind_pb += 1
        printProgressBar(ind_pb, total)

    val_set = int(((n_video * (val_split_pct * 10)) / 100) * 10)
    val_all_obj = int(((all_objects * (val_split_pct * 10)) / 100) * 10)
    if val_set == 0:
        val_set = 1

    def num_anns(elem):
        ann_dict = elem[1]
        val = 0
        for k, v in ann_dict.items():
            val = val + len(v)
        return val

    items = sorted(items, key=num_anns, reverse=True)
    train_dict = dict(items[:-val_set])

    snippets = train_val_split(train_dict)
    train = {k: v for (k, v) in snippets.items()}
    val_dict = dict(items[-val_set:])

    snippets = train_val_split(val_dict)
    val = {k: v for (k, v) in snippets.items()}

    if all_objects == 0:
        raise Exception(
            "Not enough valid data. Please ensure the labels are correct and enough data is present."
        )

    train_set = DataSets(path, train, all_objects, 5, image_name_len, ann_dict)
    val_set = DataSets(path, val, val_all_obj, 5, image_name_len)

    train_set.shuffle()
    val_set.shuffle()
    init_kwargs = {}
    train_dl = DataLoader(
        train_set,
        batch_size=batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=None,
        **init_kwargs,
    )

    valid_dl = DataLoader(
        val_set,
        batch_size=batch_size,
        num_workers=0,
        pin_memory=True,
        sampler=None,
        **init_kwargs,
    )

    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device)
    data.path = path
    data.infos = data.train_ds.infos
    data.show_batch = types.MethodType(show_batch, data)
    data._dataset_type = "ObjectTracking"
    data.train_folders = train
    data.val_folders = val
    clear_output()
    return data


def show_batch(self, rows=4, **kwargs):
    img_idxs = [random.randint(0, len(self.train_ds) - 1) for k in range(rows * 2)]
    fig, axes = plt.subplots(nrows=rows, ncols=2, squeeze=False, figsize=(20, rows * 5))
    ind = 0
    for idx in range(0, len(img_idxs), 2):
        self.train_ds.show([img_idxs[idx], img_idxs[idx + 1]], axes[ind])
        ind += 1
    pass


def _show_batch_pro(self, rows=4, **kwargs):
    img_idxs = [random.randint(0, len(self.train_ds) - 1) for k in range(rows * 2)]
    fig, axes = plt.subplots(nrows=rows, ncols=2, squeeze=False, figsize=(20, rows * 5))
    ind = 0
    for idx in range(0, len(img_idxs), 2):
        self.train_ds._show_pro([img_idxs[idx], img_idxs[idx + 1]], axes[ind])
        ind += 1
    pass



# This code is based on https://github.com/kcg2015/Vehicle-Detection-and-Tracking by kyle Guan
try:
    import numpy as np
    import cv2
    from numpy import dot
    from scipy.linalg import inv, block_diag
    from scipy.optimize import linear_sum_assignment
    import matplotlib.pyplot as plt
    from collections import deque
    import enum
except Exception as e:
    pass


class TrackStatus(enum.Enum):
    lost = 0
    searching = 8
    tracking = 16


class Track:
    """
    Creates a Track object, used to maintain the state of a track.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    id                      Required int. ID for each track initialized
    ---------------------   -------------------------------------------
    label                   Required String. label/class name of the track
    ---------------------   -------------------------------------------
    bbox                    Required list. Bounding box of the track
    ---------------------   -------------------------------------------
    mask                    Required numpy array. Mask for the tack
    =====================   ===========================================

    :return: :class:`~arcgis.learn.Track` Object
    """

    def __init__(self, id, label, bbox, mask):
        self.id = id
        self.label = label
        self.bbox = bbox
        self.score = 1
        self.status = 16
        self.mask = mask
        self.location = None
        self.age = 0


class Tracker:  # class for Kalman Filter-based tracker
    def __init__(self):
        # Initialize parametes for tracker (history)
        self.trackid = 0  # tracker's id
        self.class_name = ""
        self.score = 0.0
        self.box = []  # list to store the coordinates for a bounding box
        self.hits = 0  # number of detection matches
        self.lost_tracks = 0  # number of unmatched tracks (track loss)

        # Initialize parameters for Kalman Filtering
        # The state is the (x, y) coordinates of the detection box
        # state: [up, up_dot, left, left_dot, down, down_dot, right, right_dot]
        # or[up, up_dot, left, left_dot, height, height_dot, width, width_dot]
        self.x_state = []
        self.dt = 1.0  # time interval

        # Process matrix, assuming constant velocity model
        self.A = np.array(
            [
                [1, self.dt, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, self.dt, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, self.dt, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, self.dt],
                [0, 0, 0, 0, 0, 0, 0, 1],
            ]
        )

        # Measurement matrix, assuming we can only measure the coordinates

        self.H = np.array(
            [
                [1, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, 0],
            ]
        )

        # Initialize the state covariance
        self.state_variance = 10.0
        self.P = np.diag(self.state_variance * np.ones(8))

        # Initialize the process covariance
        self.Q_comp_mat = np.array(
            [
                [self.dt**4 / 4.0, self.dt**3 / 2.0],
                [self.dt**3 / 2.0, self.dt**2],
            ]
        )
        self.Q = block_diag(
            self.Q_comp_mat, self.Q_comp_mat, self.Q_comp_mat, self.Q_comp_mat
        )

        # Initialize the measurement covariance
        self.R_scaler = 1.0
        self.R_diag_array = self.R_scaler * np.array(
            [
                self.state_variance,
                self.state_variance,
                self.state_variance,
                self.state_variance,
            ]
        )
        self.R = np.diag(self.R_diag_array)

    def kalman_filter(self, z):
        """
        Implement the Kalman Filter, including the predict and the update stages,
        with the measurement z
        """
        x = self.x_state
        # Predict
        x = dot(self.A, x)
        self.P = dot(self.A, self.P).dot(self.A.T) + self.Q

        # Update
        S = dot(self.H, self.P).dot(self.H.T) + self.R
        K = dot(self.P, self.H.T).dot(inv(S))  # Kalman gain
        y = z - dot(self.H, x)  # residual
        x += dot(K, y)
        self.P = self.P - dot(K, self.H).dot(self.P)
        self.x_state = x.astype(int)  # convert to integer coordinates
        # (pixel values)

    def predict(self):
        """
        Implment only the predict stage. This is used for unmatched detections and
        unmatched tracks
        """
        x = self.x_state
        x = dot(self.A, x)
        self.P = dot(self.A, self.P).dot(self.A.T) + self.Q
        self.x_state = x.astype(int)


def delete_trackers(deleted_tracks, tracker_list):
    """
    Delete unused tracks from memory.

    """
    for trk in deleted_tracks:
        tracker_list.remove(trk)


def box_iou(a, b):
    """
    Helper funciton to calculate the ratio between intersection and the union of
    two boxes a and b
    a[0], a[1], a[2], a[3] <-> left, up, right, bottom
    """
    w_intsec = np.maximum(0, (np.minimum(a[2], b[2]) - np.maximum(a[0], b[0])))
    h_intsec = np.maximum(0, (np.minimum(a[3], b[3]) - np.maximum(a[1], b[1])))
    s_intsec = w_intsec * h_intsec
    s_a = (a[2] - a[0]) * (a[3] - a[1])
    s_b = (b[2] - b[0]) * (b[3] - b[1])

    return float(s_intsec) / (s_a + s_b - s_intsec)


def munkres_assignment(trackers, detections, iou_thrd):
    """
    From current list of trackers and new detections, output matched detections,
    unmatchted trackers, unmatched detections.
    """
    IOU_mat = np.zeros((len(trackers), len(detections)), dtype=np.float32)

    for t, trk in enumerate(trackers):
        for d, det in enumerate(detections):
            IOU_mat[t, d] = box_iou(trk, det)

    # Produces matches
    # Solve the maximizing the sum of IOU assignment problem using the
    # Hungarian algorithm (also known as Munkres algorithm)

    matched_idx_row, matched_idx_col = linear_sum_assignment(-IOU_mat)

    unmatched_trackers, unmatched_detections = [], []
    for t, trk in enumerate(trackers):
        if t not in matched_idx_row:
            unmatched_trackers.append(t)

    for d, det in enumerate(detections):
        if d not in matched_idx_col:
            unmatched_detections.append(d)

    matches = []

    # For creating trackers we consider any detection with an
    # overlap less than iou_thrd to signifiy the existence of
    # an untracked object
    matched_idx = np.array([matched_idx_row, matched_idx_col]).T

    for m in matched_idx:
        if IOU_mat[m[0], m[1]] < iou_thrd:
            unmatched_trackers.append(m[0])
            unmatched_detections.append(m[1])
        else:
            matches.append(m.reshape(1, 2))

    if len(matches) == 0:
        matches = np.empty((0, 2), dtype=int)
    else:
        matches = np.concatenate(matches, axis=0)

    return matches, np.array(unmatched_detections), np.array(unmatched_trackers)


def main_tracker(
    img,
    detections,
    scores,
    assignment_iou_thrd,
    vanish_frames,
    detect_frames,
    tracker_list,
    tracker_ind,
):
    """
    main_tracker function for detection and tracking
    """
    vanish_frames = vanish_frames  # no.of consecutive unmatched detection before

    detect_frames = (
        detect_frames  # no. of consecutive matches needed to establish a track
    )

    x_box = []
    z_box = detections  # measurement

    if len(tracker_list) > 0:
        for trk in tracker_list:
            x_box.append(trk.box)

    # Hungarian/Munkres Assignment
    matched, unmatched_dets, unmatched_trks = munkres_assignment(
        x_box, z_box, iou_thrd=assignment_iou_thrd
    )

    # Deal with matched detections
    if matched.size > 0:
        for trk_idx, det_idx in matched:
            z = z_box[det_idx]
            z = np.expand_dims(z, axis=0).T
            tmp_trk = tracker_list[trk_idx]
            tmp_trk.kalman_filter(z)
            xx = tmp_trk.x_state.T[0].tolist()
            xx = [xx[0], xx[2], xx[4], xx[6]]
            x_box[trk_idx] = xx
            tmp_trk.box = xx
            tmp_trk.score = scores[det_idx]
            tmp_trk.hits += 1
            tmp_trk.lost_tracks = 0

    # Deal with unmatched detections
    if len(unmatched_dets) > 0:
        for idx in unmatched_dets:
            z = z_box[idx]
            z = np.expand_dims(z, axis=0).T
            tmp_trk = Tracker()  # Create a new tracker
            x = np.array([[z[0], 0, z[1], 0, z[2], 0, z[3], 0]]).T
            tmp_trk.x_state = x
            tmp_trk.predict()
            xx = tmp_trk.x_state
            xx = xx.T[0].tolist()
            xx = [xx[0], xx[2], xx[4], xx[6]]
            tmp_trk.box = xx
            tmp_trk.score = scores[idx]
            tmp_trk.trackid = tracker_ind  # assign an ID for the tracker
            tracker_ind += 1
            tracker_list.append(tmp_trk)
            x_box.append(xx)

    # Deal with unmatched tracks
    if len(unmatched_trks) > 0:
        for trk_idx in unmatched_trks:
            tmp_trk = tracker_list[trk_idx]
            tmp_trk.lost_tracks += 1
            tmp_trk.predict()
            xx = tmp_trk.x_state
            xx = xx.T[0].tolist()
            xx = [xx[0], xx[2], xx[4], xx[6]]
            tmp_trk.box = xx
            x_box[trk_idx] = xx

    # The list of tracks to be annotated
    good_tracker_list = []
    obj_info = {}
    predictions = []
    scores = []
    labels = []
    for trk in tracker_list:
        if (trk.hits >= detect_frames) and (trk.lost_tracks <= vanish_frames):
            good_tracker_list.append(trk)
            x_cv2 = trk.box
            obj_info[trk.trackid] = (x_cv2, trk.score)

            predictions.append(
                [x_cv2[1], x_cv2[0], x_cv2[3] - x_cv2[1], x_cv2[2] - x_cv2[0]]
            )
            scores.append(trk.score)
            labels.append(trk.trackid)

    # Book keeping
    deleted_tracks = filter(lambda x: x.lost_tracks > vanish_frames, tracker_list)

    delete_trackers(deleted_tracks, tracker_list)

    return predictions, labels, scores, tracker_list, tracker_ind
